##############################################################################
# Copyright (c) 2021, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of HydraGNN and is distributed under a BSD 3-clause      #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# SPDX-License-Identifier: BSD-3-Clause                                      #
##############################################################################


def output_denormalize(y_minmax, true_values, predicted_values):
    for ihead in range(len(y_minmax)):
        ymin = y_minmax[ihead][0]
        ymax = y_minmax[ihead][1]
        for isample in range(len(predicted_values[ihead])):
            for ifeat in range(len(predicted_values[ihead][isample])):
                predicted_values[ihead][isample][ifeat] = (
                    predicted_values[ihead][isample][ifeat] * (ymax - ymin) + ymin
                )
                true_values[ihead][isample][ifeat] = (
                    true_values[ihead][isample][ifeat] * (ymax - ymin) + ymin
                )

    return true_values, predicted_values


def unscale_features_by_num_nodes(datasets_list, scaled_index_list, nodes_num_list):
    """datasets_list: list of datasets to scale back, e.g., [true_values, predicted_values];
    true_values and predicted_values[num_heads][num_samples][num_atoms]
    """
    for dataset in datasets_list:
        for scaled_index in scaled_index_list:
            head_value = dataset[scaled_index]
            for isample in range(len(nodes_num_list)):
                for iatom in range(len(head_value[isample])):
                    dataset[scaled_index][isample][iatom] *= nodes_num_list[isample]
    return datasets_list


def unscale_features_by_num_nodes_config(config, datasets_list, nodes_num_list):
    var_config = config["NeuralNetwork"]["Variables_of_interest"]
    output_names = var_config["output_names"]
    scaled_feature_index = [
        i for i in range(len(output_names)) if "_scaled_num_nodes" in output_names[i]
    ]
    if len(scaled_feature_index) > 0:
        use_denorm = var_config["denormalize_output"]
        assert use_denorm, "Cannot unscale features without 'denormalize_output'"
        datasets_list = unscale_features_by_num_nodes(
            datasets_list, scaled_feature_index, nodes_num_list
        )
    return datasets_list



##############################################################################
# Copyright (c) 2021, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of HydraGNN and is distributed under a BSD 3-clause      #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# SPDX-License-Identifier: BSD-3-Clause                                      #
##############################################################################

from enum import Enum


class AtomFeatures(Enum):
    """Class is an enum that represents features of an atom. Values paired with names of features represent column
    indexes for each feature that are used in referencing them throughout the project.
    """

    NUM_OF_PROTONS = 0
    CHARGE_DENSITY = 1
    MAGNETIC_MOMENT = 2


class StructureFeatures(Enum):
    """Class is an enum that represents features of a structure. Values paired with names of features represent column
    indexes for each feature that are used in referencing them throughout the project.
    """

    FREE_ENERGY = 0
    CHARGE_DENSITY = 1
    MAGNETIC_MOMENT = 2



import os
import pickle

import torch
from mpi4py import MPI

from .print_utils import print_distributed, log, iterate_tqdm

from hydragnn.utils.abstractbasedataset import AbstractBaseDataset
from hydragnn.preprocess import update_predicted_values, update_atom_features

import hydragnn.utils.tracer as tr


class SimplePickleDataset(AbstractBaseDataset):
    """Simple Pickle Dataset"""

    def __init__(self, basedir, label, subset=None, preload=False, var_config=None):
        """
        Parameters
        ----------
        basedir: basedir
        label: label
        subset: a list of index to subset
        """
        super().__init__()

        self.basedir = basedir
        self.label = label
        self.subset = subset
        self.preload = preload
        self.var_config = var_config
        self.input_node_features = var_config["input_node_features"]

        if self.var_config is not None:
            self.variables_type = self.var_config["type"]
            self.output_index = self.var_config["output_index"]
            self.graph_feature_dim = self.var_config["graph_feature_dims"]
            self.node_feature_dim = self.var_config["node_feature_dims"]

        fname = os.path.join(basedir, "%s-meta.pkl" % label)
        with open(fname, "rb") as f:
            self.minmax_node_feature = pickle.load(f)
            self.minmax_graph_feature = pickle.load(f)
            self.ntotal = pickle.load(f)
            self.use_subdir = pickle.load(f)
            self.nmax_persubdir = pickle.load(f)
            self.attrs = pickle.load(f)
        log("Pickle files:", self.label, self.ntotal)
        if self.attrs is None:
            self.attrs = dict()
        for k in self.attrs:
            setattr(self, k, self.attrs[k])

        if self.subset is None:
            self.subset = list(range(self.ntotal))

        if self.preload:
            for i in range(self.ntotal):
                data = self.read(i)
                self.update_data_object(data)
                self.dataset.append(data)

    def len(self):
        return len(self.subset)

    @tr.profile("get")
    def get(self, i):
        k = self.subset[i]
        if self.preload:
            return self.dataset[k]
        else:
            return self.read(k)

    def setsubset(self, subset):
        self.subset = subset

    def read(self, k):
        """
        Read from disk
        """
        fname = "%s-%d.pkl" % (self.label, k)
        dirfname = os.path.join(self.basedir, fname)
        if self.use_subdir:
            subdir = str(k // self.nmax_persubdir)
            dirfname = os.path.join(self.basedir, subdir, fname)
        with open(dirfname, "rb") as f:
            data_object = pickle.load(f)
            self.update_data_object(data_object)
        return data_object

    def setsubset(self, subset):
        self.subset = subset

    def update_data_object(self, data_object):
        if self.var_config is not None:
            update_predicted_values(
                self.variables_type,
                self.output_index,
                self.graph_feature_dim,
                self.node_feature_dim,
                data_object,
            )
            update_atom_features(self.input_node_features, data_object)


class SimplePickleWriter:
    """SimplePickleWriter class to write Torch Geometric graph data"""

    def __init__(
        self,
        dataset,
        basedir,
        label="total",
        minmax_node_feature=None,
        minmax_graph_feature=None,
        use_subdir=False,
        nmax_persubdir=10_000,
        comm=MPI.COMM_WORLD,
        attrs=dict(),
    ):
        """
        Parameters
        ----------
        dataset: locally owned dataset (should be iterable)
        basedir: basedir
        label: label
        nmax: nmax in case of subdir
        minmax_node_feature: minmax_node_feature
        minmax_graph_feature: minmax_graph_feature
        comm: MPI communicator
        """

        self.dataset = dataset
        if not isinstance(dataset, list):
            raise Exception("Unsuppored data type yet.")

        self.basedir = basedir
        self.label = label
        self.use_subdir = use_subdir
        self.nmax_persubdir = nmax_persubdir
        self.comm = comm
        self.rank = comm.Get_rank()

        self.minmax_node_feature = minmax_node_feature
        self.minmax_graph_feature = minmax_graph_feature

        ns = self.comm.allgather(len(self.dataset))
        noffset = sum(ns[: self.rank])
        ntotal = sum(ns)

        if self.rank == 0:
            if not os.path.exists(basedir):
                os.makedirs(basedir)
            fname = os.path.join(basedir, "%s-meta.pkl" % (label))
            with open(fname, "wb") as f:
                pickle.dump(self.minmax_node_feature, f)
                pickle.dump(self.minmax_graph_feature, f)
                pickle.dump(ntotal, f)
                pickle.dump(use_subdir, f)
                pickle.dump(nmax_persubdir, f)
                pickle.dump(attrs, f)
        comm.Barrier()

        if use_subdir:
            ## Create subdirs first
            subdirs = set()
            for i in range(len(self.dataset)):
                subdirs.add(str((noffset + i) // nmax_persubdir))
            for k in subdirs:
                subdir = os.path.join(basedir, k)
                os.makedirs(subdir, exist_ok=True)

        for i, data in iterate_tqdm(
            enumerate(self.dataset),
            2,
            total=len(self.dataset),
            desc="Pickle write %s" % self.label,
        ):
            fname = "%s-%d.pkl" % (label, noffset + i)
            dirfname = os.path.join(basedir, fname)
            if use_subdir:
                subdir = str((noffset + i) // nmax_persubdir)
                dirfname = os.path.join(basedir, subdir, fname)
            with open(dirfname, "wb") as f:
                pickle.dump(data, f)



##############################################################################
# Copyright (c) 2021, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of HydraGNN and is distributed under a BSD 3-clause      #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# SPDX-License-Identifier: BSD-3-Clause                                      #
##############################################################################

import torch
import torch.nn.functional as F
from torch.nn import ModuleList
from torch.nn import ReLU, Linear
from torch_geometric.nn import MFConv, BatchNorm, global_mean_pool, Sequential

from .Base import Base


class MFCStack(Base):
    def __init__(
        self,
        max_degree: int,
        *args,
        **kwargs,
    ):
        self.max_degree = max_degree

        super().__init__(*args, **kwargs)

    def get_conv(self, input_dim, output_dim):
        mfc = MFConv(
            in_channels=input_dim,
            out_channels=output_dim,
            max_degree=self.max_degree,
        )

        input_args = "x, pos, edge_index"
        conv_args = "x, edge_index"

        return Sequential(
            input_args,
            [
                (mfc, conv_args + " -> x"),
                (lambda x, pos: [x, pos], "x, pos -> x, pos"),
            ],
        )

    def __str__(self):
        return "MFCStack"



from .deterministic_graph_data import deterministic_graph_data
from .test_config import pytest_config
from .test_graphs import unittest_train_model
from .test_enthalpy import unittest_formation_enthalpy
from .test_atomicdescriptors import pytest_atomicdescriptors



##############################################################################
# Copyright (c) 2021, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of HydraGNN and is distributed under a BSD 3-clause      #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# SPDX-License-Identifier: BSD-3-Clause                                      #
##############################################################################

import sys, os, json
import pytest

import torch
import torch.distributed as dist

torch.manual_seed(97)
import shutil

import hydragnn, tests

from hydragnn.utils.config_utils import get_log_name_config
from hydragnn.utils.model import print_model
from hydragnn.utils.lsmsdataset import LSMSDataset
from hydragnn.utils.serializeddataset import SerializedWriter, SerializedDataset

from hydragnn.preprocess.load_data import split_dataset


# Main unit test function called by pytest wrappers.
def unittest_train_model_dataset_class_inheritance(
    model_type, ci_input, use_lengths, overwrite_data=False
):
    world_size, rank = hydragnn.utils.get_comm_size_and_rank()

    os.environ["SERIALIZED_DATA_PATH"] = os.getcwd()

    # Read in config settings and override model type.
    config_file = os.path.join(os.getcwd(), "tests/inputs", ci_input)
    with open(config_file, "r") as f:
        config = json.load(f)
    config["NeuralNetwork"]["Architecture"]["model_type"] = model_type

    if rank == 0:
        num_samples_tot = 500
        for dataset_name, data_path in config["Dataset"]["path"].items():
            if overwrite_data:
                shutil.rmtree(data_path)
            if not os.path.exists(data_path):
                os.makedirs(data_path)
            if dataset_name == "total":
                num_samples = num_samples_tot
            elif dataset_name == "train":
                num_samples = int(
                    num_samples_tot * config["NeuralNetwork"]["Training"]["perc_train"]
                )
            elif dataset_name == "test":
                num_samples = int(
                    num_samples_tot
                    * (1 - config["NeuralNetwork"]["Training"]["perc_train"])
                    * 0.5
                )
            elif dataset_name == "validate":
                num_samples = int(
                    num_samples_tot
                    * (1 - config["NeuralNetwork"]["Training"]["perc_train"])
                    * 0.5
                )
            if not os.listdir(data_path):
                tests.deterministic_graph_data(
                    data_path, number_configurations=num_samples
                )

    hydragnn.utils.setup_log(get_log_name_config(config))

    dirpwd = os.path.dirname(os.path.abspath(__file__))
    dirpwd = dirpwd.replace("/tests", "")
    ##################################################################################################################
    # Always initialize for multi-rank training.
    comm_size, rank = hydragnn.utils.setup_ddp()
    ##################################################################################################################

    datasetname = config["Dataset"]["name"]
    for dataset_type, raw_data_path in config["Dataset"]["path"].items():
        config["Dataset"]["path"][dataset_type] = os.path.join(dirpwd, raw_data_path)

    # In the unit test runs, it is found MFC favors graph-level features over node-level features, compared with other models;
    # hence here we decrease the loss weight coefficient for graph-level head in MFC.
    if model_type == "MFC" and ci_input == "ci_multihead.json":
        config["NeuralNetwork"]["Architecture"]["task_weights"][0] = 2

    # Only run with edge lengths for models that support them.
    if use_lengths:
        config["NeuralNetwork"]["Architecture"]["edge_features"] = ["lengths"]

    if 0 == rank:

        ## Only rank=0 is enough for pre-processing
        total = LSMSDataset(config)

        trainset, valset, testset = split_dataset(
            dataset=total,
            perc_train=config["NeuralNetwork"]["Training"]["perc_train"],
            stratify_splitting=config["Dataset"]["compositional_stratified_splitting"],
        )
        print(len(total), len(trainset), len(valset), len(testset))

        basedir = os.path.join(
            os.path.dirname(__file__), "dataset", "serialized_dataset"
        )

        SerializedWriter(
            trainset,
            basedir,
            datasetname,
            "trainset",
            minmax_node_feature=total.minmax_node_feature,
            minmax_graph_feature=total.minmax_graph_feature,
        )
        SerializedWriter(
            valset,
            basedir,
            datasetname,
            "valset",
        )
        SerializedWriter(
            testset,
            basedir,
            datasetname,
            "testset",
        )
    dist.barrier()

    print("Pickle load")
    basedir = os.path.join(os.path.dirname(__file__), "dataset", "serialized_dataset")
    trainset = SerializedDataset(basedir, datasetname, "trainset")
    valset = SerializedDataset(basedir, datasetname, "valset")
    testset = SerializedDataset(basedir, datasetname, "testset")

    ## Set minmax
    config["NeuralNetwork"]["Variables_of_interest"][
        "minmax_node_feature"
    ] = trainset.minmax_node_feature
    config["NeuralNetwork"]["Variables_of_interest"][
        "minmax_graph_feature"
    ] = trainset.minmax_graph_feature

    print(
        "trainset,valset,testset size: %d %d %d"
        % (len(trainset), len(valset), len(testset))
    )

    (train_loader, val_loader, test_loader,) = hydragnn.preprocess.create_dataloaders(
        trainset, valset, testset, config["NeuralNetwork"]["Training"]["batch_size"]
    )

    config = hydragnn.utils.update_config(config, train_loader, val_loader, test_loader)
    config["NeuralNetwork"]["Variables_of_interest"].pop("minmax_node_feature", None)
    config["NeuralNetwork"]["Variables_of_interest"].pop("minmax_graph_feature", None)

    verbosity = config["Verbosity"]["level"]
    model = hydragnn.models.create_model_config(
        config=config["NeuralNetwork"],
        verbosity=verbosity,
    )

    if 0 == rank:
        print_model(model)
    dist.barrier()

    model = hydragnn.utils.get_distributed_model(model, verbosity)

    learning_rate = config["NeuralNetwork"]["Training"]["Optimizer"]["learning_rate"]
    optimizer = torch.optim.AdamW(model.parameters(), lr=learning_rate)
    scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(
        optimizer, mode="min", factor=0.5, patience=5, min_lr=0.00001
    )

    log_name = get_log_name_config(config)
    writer = hydragnn.utils.get_summary_writer(log_name)

    if dist.is_initialized():
        dist.barrier()

    hydragnn.train.train_validate_test(
        model,
        optimizer,
        train_loader,
        val_loader,
        test_loader,
        writer,
        scheduler,
        config["NeuralNetwork"],
        log_name,
        verbosity,
        create_plots=True,
    )


# Test vector output
## FIXME: this is temporarily disabled to avoid multiple instantiations of DDP
@pytest.mark.parametrize("model_type", ["PNA"])
@pytest.mark.skip()
def pytest_train_model_vectoroutput(model_type, overwrite_data=False):
    unittest_train_model_dataset_class_inheritance(
        model_type, "ci_vectoroutput.json", True, overwrite_data
    )



##############################################################################
# Copyright (c) 2021, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of HydraGNN and is distributed under a BSD 3-clause      #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# SPDX-License-Identifier: BSD-3-Clause                                      #
##############################################################################

import sys, os, json
import pytest

import torch

torch.manual_seed(97)
import shutil

import hydragnn, tests


# Main unit test function called by pytest wrappers.
def unittest_train_model(model_type, ci_input, use_lengths, overwrite_data=False):
    world_size, rank = hydragnn.utils.get_comm_size_and_rank()

    os.environ["SERIALIZED_DATA_PATH"] = os.getcwd()

    # Read in config settings and override model type.
    config_file = os.path.join(os.getcwd(), "tests/inputs", ci_input)
    with open(config_file, "r") as f:
        config = json.load(f)
    config["NeuralNetwork"]["Architecture"]["model_type"] = model_type
    """
    to test this locally, set ci.json as
    "Dataset": {
       ...
       "path": {
               "train": "serialized_dataset/unit_test_singlehead_train.pkl",
               "test": "serialized_dataset/unit_test_singlehead_test.pkl",
               "validate": "serialized_dataset/unit_test_singlehead_validate.pkl"}
       ...
    """
    # use pkl files if exist by default
    for dataset_name in config["Dataset"]["path"].keys():
        if dataset_name == "total":
            pkl_file = (
                os.environ["SERIALIZED_DATA_PATH"]
                + "/serialized_dataset/"
                + config["Dataset"]["name"]
                + ".pkl"
            )
        else:
            pkl_file = (
                os.environ["SERIALIZED_DATA_PATH"]
                + "/serialized_dataset/"
                + config["Dataset"]["name"]
                + "_"
                + dataset_name
                + ".pkl"
            )
        if os.path.exists(pkl_file):
            config["Dataset"]["path"][dataset_name] = pkl_file

    # In the unit test runs, it is found MFC favors graph-level features over node-level features, compared with other models;
    # hence here we decrease the loss weight coefficient for graph-level head in MFC.
    if model_type == "MFC" and ci_input == "ci_multihead.json":
        config["NeuralNetwork"]["Architecture"]["task_weights"][0] = 2

    # Only run with edge lengths for models that support them.
    if use_lengths:
        config["NeuralNetwork"]["Architecture"]["edge_features"] = ["lengths"]

    if rank == 0:
        num_samples_tot = 500
        # check if serialized pickle files or folders for raw files provided
        pkl_input = False
        if list(config["Dataset"]["path"].values())[0].endswith(".pkl"):
            pkl_input = True
        # only generate new datasets, if not pkl
        if not pkl_input:
            for dataset_name, data_path in config["Dataset"]["path"].items():
                if overwrite_data:
                    shutil.rmtree(data_path)
                if not os.path.exists(data_path):
                    os.makedirs(data_path)
                if dataset_name == "total":
                    num_samples = num_samples_tot
                elif dataset_name == "train":
                    num_samples = int(
                        num_samples_tot
                        * config["NeuralNetwork"]["Training"]["perc_train"]
                    )
                elif dataset_name == "test":
                    num_samples = int(
                        num_samples_tot
                        * (1 - config["NeuralNetwork"]["Training"]["perc_train"])
                        * 0.5
                    )
                elif dataset_name == "validate":
                    num_samples = int(
                        num_samples_tot
                        * (1 - config["NeuralNetwork"]["Training"]["perc_train"])
                        * 0.5
                    )
                if not os.listdir(data_path):
                    tests.deterministic_graph_data(
                        data_path, number_configurations=num_samples
                    )

    # Since the config file uses PNA already, test the file overload here.
    # All the other models need to use the locally modified dictionary.
    if model_type == "PNA" and not use_lengths:
        hydragnn.run_training(config_file)
    else:
        hydragnn.run_training(config)

    (
        error,
        error_mse_task,
        true_values,
        predicted_values,
    ) = hydragnn.run_prediction(config)

    # Set RMSE and sample MAE error thresholds
    thresholds = {
        "SAGE": [0.20, 0.20],
        "PNA": [0.20, 0.20],
        "MFC": [0.20, 0.20],
        "GIN": [0.25, 0.20],
        "GAT": [0.60, 0.70],
        "CGCNN": [0.50, 0.40],
        "SchNet": [0.20, 0.20],
        "DimeNet": [0.50, 0.50],
        "EGNN": [0.20, 0.20],
    }
    if use_lengths and ("vector" not in ci_input):
        thresholds["CGCNN"] = [0.175, 0.175]
        thresholds["PNA"] = [0.10, 0.10]
    if use_lengths and "vector" in ci_input:
        thresholds["PNA"] = [0.2, 0.15]
    verbosity = 2

    for ihead in range(len(true_values)):
        error_head_mse = error_mse_task[ihead]
        error_str = (
            str("{:.6f}".format(error_head_mse))
            + " < "
            + str(thresholds[model_type][0])
        )
        hydragnn.utils.print_distributed(verbosity, "head: " + error_str)
        assert (
            error_head_mse < thresholds[model_type][0]
        ), "Head RMSE checking failed for " + str(ihead)

        head_true = true_values[ihead]
        head_pred = predicted_values[ihead]
        # Check individual samples
        mae = torch.nn.L1Loss()
        sample_mean_abs_error = mae(head_true, head_pred)
        error_str = (
            "{:.6f}".format(sample_mean_abs_error)
            + " < "
            + str(thresholds[model_type][1])
        )
        assert (
            sample_mean_abs_error < thresholds[model_type][1]
        ), "MAE sample checking failed!"

    # Check RMSE error
    error_str = str("{:.6f}".format(error)) + " < " + str(thresholds[model_type][0])
    hydragnn.utils.print_distributed(verbosity, "total: " + error_str)
    assert error < thresholds[model_type][0], "Total RMSE checking failed!" + str(error)


# Test across all models with both single/multihead
@pytest.mark.parametrize(
    "model_type",
    ["SAGE", "GIN", "GAT", "MFC", "PNA", "CGCNN", "SchNet", "DimeNet", "EGNN"],
)
@pytest.mark.parametrize("ci_input", ["ci.json", "ci_multihead.json"])
def pytest_train_model(model_type, ci_input, overwrite_data=False):
    unittest_train_model(model_type, ci_input, False, overwrite_data)


# Test only models
@pytest.mark.parametrize("model_type", ["PNA", "CGCNN", "SchNet", "EGNN"])
def pytest_train_model_lengths(model_type, overwrite_data=False):
    unittest_train_model(model_type, "ci.json", True, overwrite_data)


# Test across equivariant models
@pytest.mark.parametrize("model_type", ["EGNN", "SchNet"])
def pytest_train_equivariant_model(model_type, overwrite_data=False):
    unittest_train_model(model_type, "ci_equivariant.json", False, overwrite_data)


# Test vector output
@pytest.mark.parametrize("model_type", ["PNA"])
def pytest_train_model_vectoroutput(model_type, overwrite_data=False):
    unittest_train_model(model_type, "ci_vectoroutput.json", True, overwrite_data)



##############################################################################
# Copyright (c) 2021, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of HydraGNN and is distributed under a BSD 3-clause      #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# SPDX-License-Identifier: BSD-3-Clause                                      #
##############################################################################

import scipy.special
import math
import os
import shutil
import numpy as np
from tqdm import tqdm
import matplotlib.pyplot as plt


# This is LSMS specific - it assumes only one header line and only atoms following.
def read_file(path):
    with open(path, "r") as rf:
        txt = rf.readlines()
    total_energy_txt = txt[0].split()[0]

    return total_energy_txt, txt


def convert_raw_data_energy_to_gibbs(
    dir, elements_list, temperature_kelvin=0, overwrite_data=False, create_plots=True
):
    # NOTE: This works only for binary alloys

    if dir.endswith("/"):
        dir = dir[:-1]
    new_dir = dir + "_gibbs_energy/"

    if os.path.exists(new_dir):
        if overwrite_data:
            shutil.rmtree(new_dir)
    if not os.path.exists(new_dir):
        os.makedirs(new_dir)

    elements_list = sorted(elements_list)
    pure_elements_energy = dict()

    min_formation_enthalpy = float("Inf")
    max_formation_enthalpy = -float("Inf")

    # Search for the configurations with pure elements and store their total energy
    all_files = os.listdir(dir)
    for filename in tqdm(all_files):

        path = os.path.join(dir, filename)
        total_energy, txt = read_file(path)
        atoms = np.loadtxt(txt[1:])
        pure_element = np.unique(atoms[:, 0])
        if len(pure_element) == 1:
            num_atoms = atoms.shape[0]
            pure_elements_energy[pure_element[0]] = float(total_energy) / num_atoms

    assert len(pure_elements_energy) == 2, "Must have two single element files."

    num_files = len(all_files)
    total_energy_list = np.ndarray(num_files)
    linear_mixing_energy_list = np.ndarray(num_files)
    composition_list = np.ndarray(num_files)
    formation_enthalpy_list = np.ndarray(num_files)
    formation_gibbs_energy_list = np.ndarray(num_files)

    # extract formation enthalpy from total energy
    # compute thermodynamic entropy
    # compute formation gibbs energy using formation enthalpy and thermodynamic entropy
    for fn, filename in enumerate(tqdm(all_files)):

        path = os.path.join(dir, filename)
        total_energy_txt, txt = read_file(path)
        atoms = np.loadtxt(txt[1:])

        (
            composition_element1,
            total_energy,
            linear_mixing_energy,
            formation_enthalpy,
            entropy,
        ) = compute_formation_enthalpy(
            path, elements_list, pure_elements_energy, float(total_energy_txt), atoms
        )

        formation_gibbs_energy = formation_enthalpy - temperature_kelvin * entropy

        if create_plots:
            total_energy_list[fn] = total_energy
            linear_mixing_energy_list[fn] = linear_mixing_energy
            composition_list[fn] = composition_element1
            formation_enthalpy_list[fn] = formation_enthalpy
        formation_gibbs_energy_list[fn] = formation_gibbs_energy

        # Replace the total energy with mixing.
        txt[0] = txt[0].replace(total_energy_txt, str(formation_gibbs_energy))
        new_path = os.path.join(new_dir, filename)
        with open(new_path, "w") as wf:
            wf.write("".join(txt))

    min_formation_enthalpy = np.min(formation_gibbs_energy_list)
    max_formation_enthalpy = np.max(formation_gibbs_energy_list)
    print("Min formation enthalpy: ", min_formation_enthalpy)
    print("Max formation enthalpy: ", max_formation_enthalpy)

    if create_plots:
        plt.figure(0)
        plt.scatter(
            total_energy_list,
            linear_mixing_energy_list,
            edgecolor="b",
            facecolor="none",
        )
        plt.xlabel("Total energy (Rydberg)")
        plt.ylabel("Linear mixing energy (Rydberg)")
        plt.savefig("linear_mixing_energy.png")

        plt.figure(1)
        plt.scatter(
            composition_list, formation_enthalpy_list, edgecolor="b", facecolor="none"
        )
        plt.xlabel("Concentration")
        plt.ylabel("Formation enthalpy (Rydberg)")
        plt.savefig("formation_enthalpy.png")

        plt.figure(2)
        plt.scatter(
            composition_list,
            formation_gibbs_energy_list,
            edgecolor="b",
            facecolor="none",
        )
        plt.xlabel("Concentration")
        plt.ylabel("Formation Gibbs energy (Rydberg)")
        plt.savefig("formation_gibbs_energy.png")


def compute_formation_enthalpy(
    path, elements_list, pure_elements_energy, total_energy, atoms
):

    # FIXME: this currently works only for binary alloys

    elements, counts = np.unique(atoms[:, 0], return_counts=True)

    # Check all systems are in this binary.
    for e in elements:
        assert (
            e in elements_list
        ), "Sample {} contains element not present in binary considered.".format(path)
    # Fixup for the pure component cases.
    for e, elem in enumerate(elements_list):
        if elem not in elements:
            elements = np.insert(elements, e, elem)
            counts = np.insert(counts, e, 0)

    # Use first element for composition.
    num_atoms = atoms.shape[0]
    composition = counts[0] / num_atoms

    # linear_mixing_energy = energy_element1 + (energy_element2 - energy_element1) * (1-element1)
    linear_mixing_energy = (
        pure_elements_energy[elements[0]] * composition
        + pure_elements_energy[elements[1]] * (1 - composition)
    ) * num_atoms

    formation_enthalpy = total_energy - linear_mixing_energy

    # LSMS units are fixed.
    Kb_joule_per_kelvin = 1.380649 * 1e-23
    conversion_joule_rydberg = 4.5874208973812 * 1e17
    Kb_rydberg_per_kelvin = Kb_joule_per_kelvin * conversion_joule_rydberg

    # This is thermodynamic entropy, not statistical entropy
    # because we do not multiply the binomial coefficient by the probabilities
    entropy = Kb_rydberg_per_kelvin * math.log(scipy.special.comb(num_atoms, counts[0]))

    return composition, total_energy, linear_mixing_energy, formation_enthalpy, entropy



""" 
This is a tracer package to act as a wrapper to execute gptl and/or scorep.
"""

from __future__ import absolute_import
from functools import wraps
from contextlib import contextmanager

import os
import sys
from collections import OrderedDict

from abc import ABC, abstractmethod


class Tracer(ABC):
    def __init__(self, **kwargs):
        pass

    @abstractmethod
    def start(self, name):
        pass

    @abstractmethod
    def stop(self, name):
        pass

    @abstractmethod
    def enable(self):
        pass

    @abstractmethod
    def disable(self):
        pass


try:
    import gptl4py as gp

    class GPTLTracer(Tracer):
        def __init__(self, **kwargs):
            gp.initialize()

        def start(self, name):
            gp.start(name)

        def stop(self, name):
            gp.stop(name)

        def enable(self):
            gp.enable()

        def disable(self):
            gp.disable()

        def reset(self):
            gp.reset()


except:
    pass


try:
    import scorep.user as sp

    class SCOREPTracer(Tracer):
        def __init__(self, **kwargs):
            pass

        def start(self, name):
            sp.region_begin(name)

        def stop(self, name):
            sp.region_end(name)

        def enable(self):
            sp.enable_recording()

        def disable(self):
            sp.disable_recording()

        def reset(self):
            pass


except:
    pass

__tracer_list__ = dict()


def has(name):
    return name in __tracer_list__


def initialize(trlist=["GPTLTracer", "SCOREPTracer"], verbose=False, **kwargs):
    for trname in trlist:
        try:
            tr = globals()[trname](**kwargs)
            __tracer_list__[trname] = tr
        except Exception as e:
            if verbose:
                print("tracer loading error:", trname, e)
            pass


def start(name):
    for tr in __tracer_list__.values():
        tr.start(name)


def stop(name):
    for tr in __tracer_list__.values():
        tr.stop(name)


def enable():
    for tr in __tracer_list__.values():
        tr.enable()


def disable():
    for tr in __tracer_list__.values():
        tr.disable()


def reset():
    for tr in __tracer_list__.values():
        tr.reset()


def profile(x_or_func=None, *decorator_args, **decorator_kws):
    def _decorator(func):
        @wraps(func)
        def wrapper(*args, **kws):
            if "x_or_func" not in locals() or callable(x_or_func) or x_or_func is None:
                x = func.__name__
            else:
                x = x_or_func
            start(x)
            out = func(*args, **kws)
            stop(x)
            return out

        return wrapper

    return _decorator(x_or_func) if callable(x_or_func) else _decorator


@contextmanager
def timer(x):
    start(x)
    yield
    stop(x)



##############################################################################
# Copyright (c) 2021, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of HydraGNN and is distributed under a BSD 3-clause      #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# SPDX-License-Identifier: BSD-3-Clause                                      #
##############################################################################

from typing import Optional
from math import pi as PI

import torch
from torch import Tensor
from torch.nn import Identity, Linear, ReLU, Sequential
from torch_geometric.nn import Sequential as PyGSeq
from torch_geometric.nn import MessagePassing
from torch_geometric.nn.models.schnet import (
    CFConv,
    GaussianSmearing,
    RadiusInteractionGraph,
    ShiftedSoftplus,
)

from .Base import Base

from ..utils import unsorted_segment_mean


class SCFStack(Base):
    def __init__(
        self,
        num_filters: int,
        num_gaussians: list,
        radius: float,
        *args,
        max_neighbours: Optional[int] = None,
        **kwargs,
    ):
        self.radius = radius
        self.max_neighbours = max_neighbours
        self.num_filters = num_filters
        self.num_gaussians = num_gaussians

        super().__init__(*args, **kwargs)

        pass

    def _init_conv(self):

        self.distance_expansion = GaussianSmearing(0.0, self.radius, self.num_gaussians)
        self.interaction_graph = RadiusInteractionGraph(
            self.radius, self.max_neighbours
        )

        # comment on why equiv avoids last layer
        last_layer = 1 == self.num_conv_layers
        self.graph_convs.append(
            self.get_conv(self.input_dim, self.hidden_dim, last_layer)
        )
        self.feature_layers.append(Identity())
        for i in range(self.num_conv_layers - 1):
            last_layer = i == self.num_conv_layers - 2
            conv = self.get_conv(self.hidden_dim, self.hidden_dim, last_layer)
            self.graph_convs.append(conv)
            self.feature_layers.append(Identity())

    def get_conv(self, input_dim, output_dim, last_layer):
        mlp = Sequential(
            Linear(self.num_gaussians, self.num_filters),
            ShiftedSoftplus(),
            Linear(self.num_filters, self.num_filters),
        )

        interaction = CFConv(
            in_channels=input_dim,
            out_channels=output_dim,
            nn=mlp,
            num_filters=self.num_filters,
            cutoff=self.radius,
            equivariant=self.equivariance and not last_layer,
        )

        conv_args = "x, edge_index, edge_weight, edge_attr, pos"
        if self.use_edge_attr:
            input_args = "x, pos, edge_index, edge_weight, edge_attr"
            return PyGSeq(
                input_args,
                [
                    (interaction, conv_args + " -> x"),
                    (lambda x, pos: [x, pos], "x, pos -> x, pos"),
                ],
            )
        elif self.equivariance and not last_layer:
            input_args = "x, pos, batch"
            return PyGSeq(
                input_args,
                [
                    (self.interaction_graph, "pos, batch -> edge_index, edge_weight"),
                    (self.distance_expansion, "edge_weight -> edge_attr"),
                    (interaction, conv_args + " -> x, pos"),
                ],
            )
        else:
            input_args = "x, pos, batch"
            return PyGSeq(
                input_args,
                [
                    (self.interaction_graph, "pos, batch -> edge_index, edge_weight"),
                    (self.distance_expansion, "edge_weight -> edge_attr"),
                    (interaction, conv_args + " -> x"),
                    (lambda x, pos: [x, pos], "x, pos -> x, pos"),
                ],
            )

    def _conv_args(self, data):
        if (self.use_edge_attr) and (self.equivariance):
            raise Exception(
                "For SchNet if using edge attributes, then E(3)-equivariance cannot be ensured. Please disable equivariance or edge attributes."
            )
        elif self.use_edge_attr:
            edge_index = data.edge_index
            edge_weight = data.edge_attr.norm(dim=-1)

            conv_args = {
                "edge_index": edge_index,
                "edge_weight": edge_weight,
                "edge_attr": self.distance_expansion(edge_weight),
            }
        else:
            conv_args = {
                "batch": data.batch,
            }

        return conv_args

    def __str__(self):
        return "SCFStack"


class CFConv(MessagePassing):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        num_filters: int,
        nn: Sequential,
        cutoff: float,
        equivariant: bool,
    ):
        super().__init__(aggr="add")
        self.lin1 = Linear(in_channels, num_filters, bias=False)
        self.lin2 = Linear(num_filters, out_channels)
        self.nn = nn
        self.cutoff = cutoff
        self.equivariant = equivariant

        if self.equivariant:

            layer = Linear(num_filters, 1, bias=False)
            torch.nn.init.xavier_uniform_(layer.weight, gain=0.001)

            coord_mlp = []
            coord_mlp.append(Linear(num_filters, num_filters))
            coord_mlp.append(ReLU())
            coord_mlp.append(layer)
            self.coord_mlp = Sequential(*coord_mlp)

        self.reset_parameters()

    def coord_model(self, coord, edge_index, coord_diff, edge_feat):
        row, col = edge_index
        trans = coord_diff * self.coord_mlp(edge_feat)
        trans = torch.clamp(
            trans, min=-100, max=100
        )  # This is never activated but just in case it case it explosed it may save the train
        agg = unsorted_segment_mean(trans, row, num_segments=coord.size(0))
        coord = coord + agg
        return coord

    def reset_parameters(self):
        torch.nn.init.xavier_uniform_(self.lin1.weight)
        torch.nn.init.xavier_uniform_(self.lin2.weight)
        self.lin2.bias.data.fill_(0)

    def forward(
        self,
        x: Tensor,
        edge_index: Tensor,
        edge_weight: Tensor,
        edge_attr: Tensor,
        pos: Tensor,
    ) -> Tensor:
        C = 0.5 * (torch.cos(edge_weight * PI / self.cutoff) + 1.0)
        W = self.nn(edge_attr) * C.view(-1, 1)

        x = self.lin1(x)

        if self.equivariant:
            radial, coord_diff = self.coord2radial(edge_index, pos)
            pos = self.coord_model(pos, edge_index, coord_diff, W)

        x = self.propagate(edge_index, x=x, W=W)
        x = self.lin2(x)
        if self.equivariant:
            return x, pos
        else:
            return x

    def message(self, x_j: Tensor, W: Tensor) -> Tensor:
        return x_j * W

    def coord2radial(self, edge_index, coord):
        row, col = edge_index
        coord_diff = coord[row] - coord[col]
        radial = torch.sum((coord_diff) ** 2, 1).unsqueeze(1)

        norm = torch.sqrt(radial) + 1
        coord_diff = coord_diff / (norm)

        return radial, coord_diff



##############################################################################
# Copyright (c) 2021, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of HydraGNN and is distributed under a BSD 3-clause      #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# SPDX-License-Identifier: BSD-3-Clause                                      #
##############################################################################
from typing import Optional

import torch
import torch.nn as nn
from torch_geometric.nn import Sequential
from .Base import Base

from ..utils import unsorted_segment_mean


class EGCLStack(Base):
    def __init__(
        self,
        edge_attr_dim: int,
        *args,
        max_neighbours: Optional[int] = None,
        **kwargs,
    ):

        self.edge_dim = (
            0 if edge_attr_dim is None else edge_attr_dim
        )  # Must be named edge_dim to trigger use by Base
        super().__init__(*args, **kwargs)
        pass

    def _init_conv(self):
        last_layer = 1 == self.num_conv_layers
        self.graph_convs.append(
            self.get_conv(self.input_dim, self.hidden_dim, last_layer)
        )
        self.feature_layers.append(nn.Identity())
        for i in range(self.num_conv_layers - 1):
            last_layer = i == self.num_conv_layers - 2
            conv = self.get_conv(self.hidden_dim, self.hidden_dim, last_layer)
            self.graph_convs.append(conv)
            self.feature_layers.append(nn.Identity())

    def get_conv(self, input_dim, output_dim, last_layer=False):
        egcl = E_GCL(
            input_channels=input_dim,
            output_channels=output_dim,
            hidden_channels=self.hidden_dim,
            edge_attr_dim=self.edge_dim,
            equivariant=self.equivariance and not last_layer,
        )

        if self.equivariance and not last_layer:
            return Sequential(
                "x, pos, edge_index, edge_attr",
                [
                    (egcl, "x, pos, edge_index, edge_attr -> x, pos"),
                ],
            )
        else:
            return Sequential(
                "x, pos, edge_index, edge_attr",
                [
                    (egcl, "x, pos, edge_index, edge_attr -> x"),
                    (lambda x, pos: [x, pos], "x, pos -> x, pos"),
                ],
            )

    def _conv_args(self, data):
        if self.edge_dim > 0:
            conv_args = {
                "edge_index": data.edge_index,
                "edge_attr": data.edge_attr,
            }
        else:
            conv_args = {
                "edge_index": data.edge_index,
                "edge_attr": None,
            }

        return conv_args

    def __str__(self):
        return "EGCLStack"


"""
EGNN
=====

E(n) equivariant graph neural network as
a message passing neural network. The
model uses positional data only to ensure
that the message passing component is
equivariant.

In particular this message passing layer
relies on the angle formed by the triplet
of incomming and outgoing messages.

The three key components of this network are
outlined below. In particular, the convolutional
network that is used for the message passing
the triplet function that generates to/from
information for angular values, and finally
the radial basis embedding that is used to
include radial basis information.

"""


class E_GCL(nn.Module):
    def __init__(
        self,
        input_channels,
        output_channels,
        hidden_channels,
        edge_attr_dim=0,
        nodes_attr_dim=0,
        act_fn=nn.ReLU(),
        recurrent=False,
        coords_weight=1.0,
        attention=False,
        clamp=False,
        norm_diff=True,
        tanh=True,
        equivariant=False,
    ) -> None:
        super(E_GCL, self).__init__()
        input_edge = input_channels * 2
        self.coords_weight = coords_weight
        self.recurrent = recurrent
        self.attention = attention
        self.norm_diff = norm_diff
        self.tanh = tanh
        edge_coords_nf = 1
        self.equivariant = equivariant
        self.edge_attr_dim = edge_attr_dim

        self.edge_mlp = nn.Sequential(
            nn.Linear(input_edge + edge_coords_nf + edge_attr_dim, hidden_channels),
            act_fn,
            nn.Linear(hidden_channels, hidden_channels),
            act_fn,
        )

        self.node_mlp = nn.Sequential(
            nn.Linear(
                hidden_channels + input_channels + nodes_attr_dim, hidden_channels
            ),
            act_fn,
            nn.Linear(hidden_channels, output_channels),
        )

        self.clamp = clamp

        if self.equivariant:

            layer = nn.Linear(hidden_channels, 1, bias=False)
            torch.nn.init.xavier_uniform_(layer.weight, gain=0.001)

            coord_mlp = []
            coord_mlp.append(nn.Linear(hidden_channels, hidden_channels))
            coord_mlp.append(act_fn)
            coord_mlp.append(layer)
            if self.tanh:
                coord_mlp.append(nn.Tanh())
                self.coords_range = nn.Parameter(torch.ones(1)) * 3
            self.coord_mlp = nn.Sequential(*coord_mlp)

        if self.attention:
            self.att_mlp = nn.Sequential(nn.Linear(hidden_channels, 1), nn.Sigmoid())

        self.act_fn = act_fn
        pass

    def edge_model(self, source, target, radial, edge_attr):
        if edge_attr is None:  # Unused.
            out = torch.cat([source, target, radial], dim=1)
        else:
            out = torch.cat([source, target, radial, edge_attr], dim=1)
        out = self.edge_mlp(out)
        if self.attention:
            att_val = self.att_mlp(out)
            out = out * att_val
        return out

    def node_model(self, x, edge_index, edge_attr, node_attr):
        row, col = edge_index
        agg = unsorted_segment_sum(edge_attr, row, num_segments=x.size(0))
        if node_attr is not None:
            agg = torch.cat([x, agg, node_attr], dim=1)
        else:
            agg = torch.cat([x, agg], dim=1)
        out = self.node_mlp(agg)
        if self.recurrent:
            out = x + out
        return out, agg

    def coord_model(self, coord, edge_index, coord_diff, edge_feat):
        row, col = edge_index
        trans = coord_diff * self.coord_mlp(edge_feat)
        trans = torch.clamp(
            trans, min=-100, max=100
        )  # This is never activated but just in case it case it explosed it may save the train
        agg = unsorted_segment_mean(trans, row, num_segments=coord.size(0))
        coord = coord + agg * self.coords_weight
        return coord

    def coord2radial(self, edge_index, coord):
        row, col = edge_index
        coord_diff = coord[row] - coord[col]
        radial = torch.sum((coord_diff) ** 2, 1).unsqueeze(1)

        if self.norm_diff:
            norm = torch.sqrt(radial) + 1
            coord_diff = coord_diff / (norm)

        return radial, coord_diff

    def forward(self, x, coord, edge_index, edge_attr, node_attr=None):
        row, col = edge_index
        radial, coord_diff = self.coord2radial(edge_index, coord)
        # Message Passing
        edge_feat = self.edge_model(x[row], x[col], radial, edge_attr)
        if self.equivariant:
            coord = self.coord_model(coord, edge_index, coord_diff, edge_feat)
        x, agg = self.node_model(x, edge_index, edge_feat, node_attr)
        if self.equivariant:
            return x, coord
        else:
            return x


def unsorted_segment_sum(data, segment_ids, num_segments):
    """Custom PyTorch op to replicate TensorFlow's `unsorted_segment_sum`."""
    result_shape = (num_segments, data.size(1))
    result = data.new_full(result_shape, 0)  # Init empty result tensor.
    segment_ids = segment_ids.unsqueeze(-1).expand(-1, data.size(1))
    result.scatter_add_(0, segment_ids, data)
    return result



##############################################################################
# Copyright (c) 2021, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of HydraGNN and is distributed under a BSD 3-clause      #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# SPDX-License-Identifier: BSD-3-Clause                                      #
##############################################################################

from tqdm import tqdm

import torch.distributed as dist

import logging
import os


"""
Verbosity options for printing
0 - > nothing
1 -> master prints the basic
2 -> Master prints everything, progression bars included
3 -> all MPI processes print the basic
4 -> all MPI processes print the basic, , progression bars included
"""


def print_nothing(*args):
    pass


def print_master(*args):
    log(*args, rank=0)


def print_all_processes(*args):
    log(*args)


switcher = {
    0: print_nothing,
    1: print_master,
    2: print_master,
    3: print_all_processes,
    4: print_all_processes,
}


def print_distributed(verbosity_level, *args):
    print_verbose = switcher.get(verbosity_level)
    return print_verbose(*args)


def iterate_tqdm(iterator, verbosity_level, *args, **kwargs):
    if (0 == dist.get_rank() and 2 == verbosity_level) or 4 == verbosity_level:
        return tqdm(iterator, *args, **kwargs)
    else:
        return iterator


def setup_log(prefix):
    """
    Setup logging to print messages for both screen and file.
    """
    from .distributed import init_comm_size_and_rank

    world_size, world_rank = init_comm_size_and_rank()

    fmt = "%d: %%(message)s" % (world_rank)
    logFormatter = logging.Formatter(fmt)

    logger = logging.getLogger("hydragnn")
    logger.propagate = False
    logger.setLevel(logging.DEBUG)

    os.makedirs("./logs/%s" % prefix, exist_ok=True)
    fname = "./logs/%s/run.log" % (prefix)
    fileHandler = logging.FileHandler(fname)

    fileHandler.setFormatter(logFormatter)
    logger.addHandler(fileHandler)

    consoleHandler = logging.StreamHandler()
    consoleHandler.setFormatter(logFormatter)
    logger.addHandler(consoleHandler)


def log(*args, sep=" ", rank=None):
    """
    Helper function to print/log messages.
    rank parameter is to limit which rank should print. if rank is None, all processes print.
    """
    logger = logging.getLogger("hydragnn")

    if rank is None:
        logger.info(sep.join(map(str, args)))
    else:
        from .distributed import init_comm_size_and_rank

        world_size, world_rank = init_comm_size_and_rank()
        if rank == world_rank:
            logger.info(sep.join(map(str, args)))



import os
import numpy as np

from torch import tensor
from torch_geometric.data import Data
from hydragnn.utils.abstractrawdataset import AbstractRawDataset

from ase.io.cfg import read_cfg


class CFGDataset(AbstractRawDataset):
    def __init__(self, config, dist=False, sampling=None):
        super().__init__(config, dist, sampling)

    def transform_input_to_data_object_base(self, filepath):
        data_object = self.__transform_CFG_input_to_data_object_base(filepath=filepath)
        return data_object

    def __transform_CFG_input_to_data_object_base(self, filepath):
        """Transforms lines of strings read from the raw data CFG file to Data object and returns it.

        Parameters
        ----------
        lines:
          content of data file with all the graph information
        Returns
        ----------
        Data
            Data object representing structure of a graph sample.
        """

        if filepath.endswith(".cfg"):

            data_object = self.__transform_ASE_object_to_data_object(filepath)

            return data_object

        else:
            return None

    def __transform_ASE_object_to_data_object(self, filepath):

        # FIXME:
        #  this still assumes bulk modulus is specific to the CFG format.
        #  To deal with multiple files across formats, one should generalize this function
        #  by moving the reading of the .bulk file in a standalone routine.
        #  Morevoer, this approach assumes tha there is only one global feature to look at,
        #  and that this global feature is specicially retrieveable in a file with the string *bulk* inside.

        ase_object = read_cfg(filepath)

        data_object = Data()

        data_object.supercell_size = tensor(ase_object.cell.array).float()
        data_object.pos = tensor(ase_object.arrays["positions"]).float()
        proton_numbers = np.expand_dims(ase_object.arrays["numbers"], axis=1)
        masses = np.expand_dims(ase_object.arrays["masses"], axis=1)
        c_peratom = np.expand_dims(ase_object.arrays["c_peratom"], axis=1)
        fx = np.expand_dims(ase_object.arrays["fx"], axis=1)
        fy = np.expand_dims(ase_object.arrays["fy"], axis=1)
        fz = np.expand_dims(ase_object.arrays["fz"], axis=1)
        node_feature_matrix = np.concatenate(
            (proton_numbers, masses, c_peratom, fx, fy, fz), axis=1
        )
        data_object.x = tensor(node_feature_matrix).float()

        filename_without_extension = os.path.splitext(filepath)[0]

        if os.path.exists(os.path.join(filename_without_extension + ".bulk")):
            filename_bulk = os.path.join(filename_without_extension + ".bulk")
            f = open(filename_bulk, "r", encoding="utf-8")
            lines = f.readlines()
            graph_feat = lines[0].split(None, 2)
            g_feature = []
            # collect graph features
            for item in range(len(self.graph_feature_dim)):
                for icomp in range(self.graph_feature_dim[item]):
                    it_comp = self.graph_feature_col[item] + icomp
                    g_feature.append(float(graph_feat[it_comp].strip()))
            data_object.y = tensor(g_feature)

        return data_object



import torch
import contextlib
from unittest.mock import MagicMock
from torch.profiler import profile, record_function, ProfilerActivity

from hydragnn.utils.distributed import get_device_name


class Profiler(torch.profiler.profile):
    def __init__(self, prefix="", enable=False, target_epoch=0):
        self.prefix = prefix
        self.enable = enable
        self.target_epoch = target_epoch
        self.current_epoch = -1
        self.done = False

        # FIXME: can remove check for cuda when torch-1.9 is requireds
        activities = [ProfilerActivity.CPU]
        device = get_device_name()
        if "cuda" in device:
            activities.append(ProfilerActivity.CUDA)
        super(Profiler, self).__init__(
            activities=activities,
            schedule=torch.profiler.schedule(wait=5, warmup=3, active=3, repeat=1),
            on_trace_ready=self.trace_handler,
            record_shapes=True,
            with_stack=True,
        )

        self.null_profiler = contextlib.nullcontext(MagicMock(name="step"))

    def setup(self, config):
        def tryget(key, default):
            val = default
            try:
                val = config[key]
            except:
                pass
            return val

        self.enable = tryget("enable", 0) == 1
        self.target_epoch = tryget("target_epoch", 0)

    def trace_handler(self, p):
        print(
            "Total number of profiled events: %d at epoch %d"
            % (len(p.events()), self.target_epoch)
        )
        torch.profiler.tensorboard_trace_handler(self.prefix)(p)
        self.done = True

    def set_current_epoch(self, current_epoch):
        self.current_epoch = current_epoch

    def __enter__(self):
        if self.enable and (self.target_epoch == self.current_epoch):
            return super(Profiler, self).__enter__()
        else:
            return self.null_profiler.__enter__()

    def __exit__(self, type, value, traceback):
        if self.enable and (self.target_epoch == self.current_epoch):
            super(Profiler, self).__exit__(type, value, traceback)

    def step(self):
        if self.enable:
            super(Profiler, self).step()

    def reset(self):
        super(Profiler, self).step_num = 0



##############################################################################
# Copyright (c) 2021, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of HydraGNN and is distributed under a BSD 3-clause      #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# SPDX-License-Identifier: BSD-3-Clause                                      #
##############################################################################

from typing import Callable, Optional, Tuple
from torch_geometric.typing import SparseTensor

import torch
from torch import Tensor
from torch.nn import Identity, SiLU

from torch_geometric.nn import Linear, Sequential
from torch_geometric.nn.models.dimenet import (
    BesselBasisLayer,
    EmbeddingBlock,
    InteractionPPBlock,
    OutputPPBlock,
    SphericalBasisLayer,
)
from torch_geometric.utils import scatter

from .Base import Base


class DIMEStack(Base):
    """
    Generates angles, distances, to/from indices, radial basis
    functions and spherical basis functions for learning.
    """

    def __init__(
        self,
        basis_emb_size,
        envelope_exponent,
        int_emb_size,
        out_emb_size,
        num_after_skip,
        num_before_skip,
        num_radial,
        num_spherical,
        radius,
        *args,
        max_neighbours: Optional[int] = None,
        **kwargs
    ):
        self.basis_emb_size = basis_emb_size
        self.int_emb_size = int_emb_size
        self.out_emb_size = out_emb_size
        self.num_radial = num_radial
        self.num_spherical = num_spherical
        self.num_before_skip = num_before_skip
        self.num_after_skip = num_after_skip
        self.radius = radius

        super().__init__(*args, **kwargs)

        self.rbf = BesselBasisLayer(num_radial, radius, envelope_exponent)
        self.sbf = SphericalBasisLayer(
            num_spherical, num_radial, radius, envelope_exponent
        )

        pass

    def _init_conv(self):
        self.graph_convs.append(self.get_conv(self.input_dim, self.hidden_dim))
        self.feature_layers.append(Identity())
        for _ in range(self.num_conv_layers - 1):
            conv = self.get_conv(self.hidden_dim, self.hidden_dim)
            self.graph_convs.append(conv)
            self.feature_layers.append(Identity())

    def get_conv(self, input_dim, output_dim):
        hidden_dim = output_dim if input_dim == 1 else input_dim
        assert (
            hidden_dim > 1
        ), "DimeNet requires more than one hidden dimension between input_dim and output_dim."
        lin = Linear(input_dim, hidden_dim)
        emb = HydraEmbeddingBlock(
            num_radial=self.num_radial, hidden_channels=hidden_dim, act=SiLU()
        )
        inter = InteractionPPBlock(
            hidden_channels=hidden_dim,
            int_emb_size=self.int_emb_size,
            basis_emb_size=self.basis_emb_size,
            num_spherical=self.num_spherical,
            num_radial=self.num_radial,
            num_before_skip=self.num_before_skip,
            num_after_skip=self.num_after_skip,
            act=SiLU(),
        )
        dec = OutputPPBlock(
            num_radial=self.num_radial,
            hidden_channels=hidden_dim,
            out_emb_channels=self.out_emb_size,
            out_channels=output_dim,
            num_layers=1,
            act=SiLU(),
        )
        return Sequential(
            "x, pos, rbf, sbf, i, j, idx_kj, idx_ji",
            [
                (lin, "x -> x"),
                (emb, "x, rbf, i, j -> x1"),
                (inter, "x1, rbf, sbf, idx_kj, idx_ji -> x2"),
                (dec, "x2, rbf, i -> c"),
                (lambda x, pos: [x, pos], "c, pos -> c, pos"),
            ],
        )

    def _conv_args(self, data):
        assert (
            data.pos is not None
        ), "DimeNet requires node positions (data.pos) to be set."
        i, j, idx_i, idx_j, idx_k, idx_kj, idx_ji = triplets(
            data.edge_index, num_nodes=data.x.size(0)
        )
        dist = (data.pos[i] - data.pos[j]).pow(2).sum(dim=-1).sqrt()

        # Calculate angles.
        pos_i = data.pos[idx_i]
        pos_ji, pos_ki = data.pos[idx_j] - pos_i, data.pos[idx_k] - pos_i
        a = (pos_ji * pos_ki).sum(dim=-1)
        b = torch.cross(pos_ji, pos_ki).norm(dim=-1)
        angle = torch.atan2(b, a)

        rbf = self.rbf(dist)
        sbf = self.sbf(dist, angle, idx_kj)

        conv_args = {
            "rbf": rbf,
            "sbf": sbf,
            "i": i,
            "j": j,
            "idx_kj": idx_kj,
            "idx_ji": idx_ji,
        }

        return conv_args


"""
PyG Adapted Codes
------------------
The following code is adapted from
https://github.com/pyg-team/pytorch_geometric/blob/master/torch_geometric/nn/models/dimenet.py

"""


def triplets(
    edge_index: Tensor,
    num_nodes: int,
) -> Tuple[Tensor, Tensor, Tensor, Tensor, Tensor, Tensor, Tensor]:
    row, col = edge_index  # j->i

    value = torch.arange(row.size(0), device=row.device)
    adj_t = SparseTensor(
        row=col, col=row, value=value, sparse_sizes=(num_nodes, num_nodes)
    )
    adj_t_row = adj_t[row]
    num_triplets = adj_t_row.set_value(None).sum(dim=1).to(torch.long)

    # Node indices (k->j->i) for triplets.
    idx_i = col.repeat_interleave(num_triplets)
    idx_j = row.repeat_interleave(num_triplets)
    idx_k = adj_t_row.storage.col()
    mask = idx_i != idx_k  # Remove i == k triplets.
    idx_i, idx_j, idx_k = idx_i[mask], idx_j[mask], idx_k[mask]

    # Edge indices (k-j, j->i) for triplets.
    idx_kj = adj_t_row.storage.value()[mask]
    idx_ji = adj_t_row.storage.row()[mask]

    return col, row, idx_i, idx_j, idx_k, idx_kj, idx_ji


class HydraEmbeddingBlock(EmbeddingBlock):
    def __init__(self, num_radial: int, hidden_channels: int, act: Callable):
        super().__init__(
            num_radial=num_radial, hidden_channels=hidden_channels, act=act
        )
        del self.emb  # Atomic embeddings are handled by Hydra.
        self.reset_parameters()

    def reset_parameters(self):
        # self.emb.weight.data.uniform_(-sqrt(3), sqrt(3))
        self.lin_rbf.reset_parameters()
        self.lin.reset_parameters()

    def forward(self, x: Tensor, rbf: Tensor, i: Tensor, j: Tensor) -> Tensor:
        # x = self.emb(x)
        rbf = self.act(self.lin_rbf(rbf))
        return self.act(self.lin(torch.cat([x[i], x[j], rbf], dim=-1)))



##############################################################################
# Copyright (c) 2021, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of HydraGNN and is distributed under a BSD 3-clause      #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# SPDX-License-Identifier: BSD-3-Clause                                      #
##############################################################################

import os, json
import pytest


@pytest.mark.parametrize("config_file", ["lsms/lsms.json"])
@pytest.mark.mpi_skip()
def pytest_config(config_file):

    config_file = os.path.join("examples", config_file)
    with open(config_file, "r") as f:
        config = json.load(f)

    expected = {
        "Dataset": [
            "name",
            "path",
            "format",
            "num_nodes",
            "node_features",
            "graph_features",
        ],
        "NeuralNetwork": ["Architecture", "Variables_of_interest", "Training"],
    }

    for category in expected.keys():
        assert category in config, "Missing required input category"

        for input in category:
            assert input in category, "Missing required input"



from mpi4py import MPI
import time
import os
import glob

from .print_utils import print_distributed, log, iterate_tqdm

import numpy as np

try:
    import adios2 as ad2
except ImportError:
    pass

import torch_geometric.data
import torch

from multiprocessing.shared_memory import SharedMemory

try:
    import pyddstore as dds
except ImportError:
    pass

import hydragnn.utils.tracer as tr

from hydragnn.utils.abstractbasedataset import AbstractBaseDataset
from hydragnn.utils import nsplit


class AdiosWriter:
    """Adios class to write Torch Geometric graph data"""

    def __init__(self, filename, comm):
        """
        Parameters
        ----------
        filename: str
            adios filename
        comm: MPI_comm
            MPI communicator to use for Adios parallel writing
        """
        self.filename = filename
        self.comm = comm
        self.rank = comm.Get_rank()
        self.size = comm.Get_size()

        self.dataset = dict()
        self.attributes = dict()
        self.adios = ad2.ADIOS()
        self.io = self.adios.DeclareIO(self.filename)

    def add_global(self, vname, arr):
        """
        Add attribute to be written in adios file

        Parameters
        ----------
        vname: str
            attribute name
        arr: numpy array
            attribute value to be written
        """
        self.attributes[vname] = arr

    def add(self, label, data):
        """
        Add labeled data to be written in adios file

        Parameters
        ----------
        label: str
            label name
        data: PyG data or list of PyG data
            PyG data to be saved
        """
        if label not in self.dataset:
            self.dataset[label] = list()

        if isinstance(data, list):
            self.dataset[label].extend(data)
        elif isinstance(data, torch_geometric.data.Data):
            self.dataset[label].append(data)
        elif isinstance(data, BaseDataset):
            self.dataset[label] = data
        else:
            raise Exception("Unsuppored data type yet.")

    def save(self):
        """
        Save data into an Adios file
        """
        t0 = time.time()
        log("Adios saving:", self.filename)
        self.writer = self.io.Open(self.filename, ad2.Mode.Write, self.comm)
        total_ns = 0
        for label in self.dataset:
            if len(self.dataset[label]) == 0:
                ## If there is no data to save, simply do empty operations as follows.
                ## This process will call multiple allgather in a sequential order
                ns = self.comm.allgather(len(self.dataset[label]))

                keys_list = self.comm.allgather([])
                for keys in keys_list:
                    if len(keys) > 0:
                        break

                for k in keys:
                    shape_list = self.comm.allgather([])

                continue
            ns = self.comm.allgather(len(self.dataset[label]))
            ns_offset = sum(ns[: self.rank])

            self.io.DefineAttribute("%s/ndata" % label, np.array(sum(ns)))
            total_ns += sum(ns)

            if len(self.dataset[label]) > 0:
                data = self.dataset[label][0]
                self.io.DefineAttribute("%s/keys" % label, data.keys)
                keys = sorted(data.keys)
                self.comm.allgather(keys)

            for k in keys:
                arr_list = [data[k].cpu().numpy() for data in self.dataset[label]]
                m0 = np.min([x.shape for x in arr_list], axis=0)
                m1 = np.max([x.shape for x in arr_list], axis=0)
                vdims = list()
                for i in range(len(m0)):
                    if m0[i] != m1[i]:
                        vdims.append(i)
                ## We can handle only one varying dimension.
                assert len(vdims) < 2
                vdim = 0
                if len(vdims) > 0:
                    vdim = vdims[0]
                val = np.concatenate(arr_list, axis=vdim)
                assert val.data.contiguous

                shape_list = self.comm.allgather(list(val.shape))
                offset = [
                    0,
                ] * len(val.shape)
                for i in range(self.rank):
                    if shape_list[i]:
                        offset[vdim] += shape_list[i][vdim]
                global_shape = shape_list[0]
                for i in range(1, self.size):
                    if shape_list[i]:
                        global_shape[vdim] += shape_list[i][vdim]
                # log ("k,val shape", k, global_shape, offset, val.shape)
                var = self.io.DefineVariable(
                    "%s/%s" % (label, k),
                    val,
                    global_shape,
                    offset,
                    val.shape,
                    ad2.ConstantDims,
                )
                self.writer.Put(var, val, ad2.Mode.Sync)

                self.io.DefineAttribute(
                    "%s/%s/variable_dim" % (label, k), np.array(vdim)
                )

                vcount = np.array([x.shape[vdim] for x in arr_list])
                assert len(vcount) == len(self.dataset[label])

                offset_arr = np.zeros_like(vcount)
                offset_arr[1:] = np.cumsum(vcount)[:-1]
                offset_arr += offset[vdim]

                var = self.io.DefineVariable(
                    "%s/%s/variable_count" % (label, k),
                    vcount,
                    [
                        sum(ns),
                    ],
                    [
                        ns_offset,
                    ],
                    [
                        len(vcount),
                    ],
                    ad2.ConstantDims,
                )
                self.writer.Put(var, vcount, ad2.Mode.Sync)

                var = self.io.DefineVariable(
                    "%s/%s/variable_offset" % (label, k),
                    offset_arr,
                    [
                        sum(ns),
                    ],
                    [
                        ns_offset,
                    ],
                    [
                        len(vcount),
                    ],
                    ad2.ConstantDims,
                )
                self.writer.Put(var, offset_arr, ad2.Mode.Sync)

        self.io.DefineAttribute("total_ndata", np.array(total_ns))
        for vname in self.attributes:
            self.io.DefineAttribute(vname, self.attributes[vname])

        self.writer.Close()
        t1 = time.time()
        log("Adios saving time (sec): ", (t1 - t0))


class AdiosDataset(AbstractBaseDataset):
    """Adios dataset class"""

    def __init__(
        self,
        filename,
        label,
        comm,
        preload=False,
        shmem=False,
        enable_cache=False,
        ddstore=False,
        ddstore_width=None,
    ):
        """
        Parameters
        ----------
        filename: str
            adios filename
        label: str
            data label to load, such as trainset, testing, and valset
        comm: MPI_comm
            MPI communicator
        preload: bool, optional
            Option to preload all the dataset into a memory
        shmem: bool, optional
            Option to use shmem to share data between processes in the same node
        enable_cache: bool, optional
            Option to cache data object which was already read
        ddstore: bool, optional
            Option to use Distributed Data Store
        """
        t0 = time.time()
        self.filename = filename
        self.label = label
        self.comm = comm
        self.rank = self.comm.Get_rank()
        self.comm_size = self.comm.Get_size()

        self.nrank_per_node = self.comm.Get_size()
        if os.getenv("OMPI_COMM_WORLD_LOCAL_SIZE"):
            ## Summit
            self.nrank_per_node = int(os.environ["OMPI_COMM_WORLD_LOCAL_SIZE"])
        elif os.getenv("SLURM_NTASKS_PER_NODE"):
            ## Perlmutter
            self.nrank_per_node = int(os.environ["SLURM_NTASKS_PER_NODE"])
        elif os.getenv("SLURM_TASKS_PER_NODE"):
            ## Crusher
            self.nrank_per_node = int(os.environ["SLURM_TASKS_PER_NODE"].split("(")[0])

        self.data = dict()
        self.preload = preload
        ## Preflight option: This is experimental.
        ## Get the index of data to load first and call "populate" to load data all together
        self.preflight = False
        self.preflight_list = list()
        self.shmem = shmem
        self.smm = None
        if self.shmem:
            self.shm = dict()
            self.local_comm = self.comm.Split(
                self.rank // self.nrank_per_node, self.rank
            )
            self.local_rank = self.local_comm.Get_rank()
            log("local_rank", self.local_rank)

        self.enable_cache = enable_cache
        self.cache = dict()
        self.ddstore = None
        self.ddstore = ddstore
        self.ddstore_width = (
            ddstore_width if ddstore_width is not None else self.comm_size
        )
        if self.ddstore:
            self.ddstore_comm = self.comm.Split(
                self.rank // self.ddstore_width, self.rank
            )
            self.ddstore_comm_rank = self.ddstore_comm.Get_rank()
            self.ddstore_comm_size = self.ddstore_comm.Get_size()
            log(
                "ddstore_comm_rank,ddstore_comm_size",
                self.ddstore_comm_rank,
                self.ddstore_comm_size,
            )
            self.ddstore = dds.PyDDStore(self.ddstore_comm)
        log("Adios reading:", self.filename)
        with ad2.open(self.filename, "r", MPI.COMM_SELF) as f:
            f.__next__()
            self.vars = f.available_variables()
            self.attrs = f.available_attributes()
            self.keys = f.read_attribute_string("%s/keys" % label)
            self.ndata = f.read_attribute("%s/ndata" % label).item()
            if "minmax_graph_feature" in self.attrs:
                self.minmax_graph_feature = f.read_attribute(
                    "minmax_graph_feature"
                ).reshape((2, -1))
            if "minmax_node_feature" in self.attrs:
                self.minmax_node_feature = f.read_attribute(
                    "minmax_node_feature"
                ).reshape((2, -1))
            if "pna_deg" in self.attrs:
                self.pna_deg = f.read_attribute("pna_deg")

            self.variable_count = dict()
            self.variable_offset = dict()
            self.variable_dim = dict()

            nbytes = 0
            for k in self.keys:
                self.variable_count[k] = f.read("%s/%s/variable_count" % (label, k))
                self.variable_offset[k] = f.read("%s/%s/variable_offset" % (label, k))
                self.variable_dim[k] = f.read_attribute(
                    "%s/%s/variable_dim" % (label, k)
                ).item()
                if self.preload:
                    ## load full data first
                    self.data[k] = f.read("%s/%s" % (label, k))
                elif self.shmem:
                    if self.local_rank == 0:
                        adios = ad2.ADIOS()
                        io = adios.DeclareIO("ogb_read")
                        reader = io.Open(self.filename, ad2.Mode.Read, MPI.COMM_SELF)
                        var = io.InquireVariable("%s/%s" % (label, k))

                        if var.Type() == "double":
                            dtype = np.float64
                        elif var.Type() == "float":
                            dtype = np.float32
                        elif var.Type() == "int32_t":
                            dtype = np.int32
                        elif var.Type() == "int64_t":
                            dtype = np.int64
                        else:
                            raise ValueError(var.Type())

                        nbytes = np.prod(var.Shape()) * np.dtype(dtype).itemsize
                        self.shm[k] = SharedMemory(create=True, size=nbytes)
                        arr = np.ndarray(
                            var.Shape(), dtype=dtype, buffer=self.shm[k].buf
                        )
                        reader.Get(var, arr, ad2.Mode.Sync)
                        reader.Close()
                        self.data[k] = arr
                        self.local_comm.bcast(self.shm[k].name, root=0)
                    else:
                        name = None
                        name = self.local_comm.bcast(name, root=0)
                        self.shm[k] = SharedMemory(name=name, create=False)

                        shape = self.vars["%s/%s" % (self.label, k)]["Shape"]
                        ishape = [int(x.strip(",")) for x in shape.strip().split()]

                        vartype = self.vars["%s/%s" % (self.label, k)]["Type"]
                        if vartype == "double":
                            dtype = np.float64
                        elif vartype == "float":
                            dtype = np.float32
                        elif vartype == "int32_t":
                            dtype = np.int32
                        elif vartype == "int64_t":
                            dtype = np.int64
                        else:
                            raise ValueError(vartype)

                        arr = np.ndarray(ishape, dtype=dtype, buffer=self.shm[k].buf)
                        self.data[k] = arr
                elif self.ddstore:
                    ## Calculate local portion
                    shape = self.vars["%s/%s" % (self.label, k)]["Shape"]
                    ishape = [int(x.strip(",")) for x in shape.strip().split()]
                    start = [
                        0,
                    ] * len(ishape)
                    count = ishape
                    vdim = self.variable_dim[k]

                    rx = list(nsplit(self.variable_count[k], self.ddstore_comm_size))
                    start[vdim] = sum([sum(x) for x in rx[: self.ddstore_comm_rank]])
                    count[vdim] = sum(rx[self.ddstore_comm_rank])

                    # Read only local portion
                    vname = "%s/%s" % (label, k)
                    self.data[k] = f.read(vname, start, count)
                    if vdim > 0:
                        self.data[k] = np.moveaxis(self.data[k], vdim, 0)
                        self.data[k] = np.ascontiguousarray(self.data[k])
                    self.ddstore.add(vname, self.data[k])
                    log(
                        "DDStore add:",
                        (
                            vname,
                            start,
                            count,
                            vdim,
                            self.data[k].dtype,
                            self.data[k].shape,
                            self.data[k].sum(),
                        ),
                    )
                    nbytes += self.data[k].size * self.data[k].itemsize
            t2 = time.time()
            log("Adios reading time (sec): ", (t2 - t0))
            if self.ddstore:
                log("DDStore total (GB):", nbytes / 1024 / 1024 / 1024)

        t1 = time.time()
        log("Data loading time (sec): ", (t1 - t0))

        if not self.preload and not self.shmem:
            self.f = ad2.open(self.filename, "r", MPI.COMM_SELF)
            self.f.__next__()

    def len(self):
        """
        Return the total size of dataset
        """
        return self.ndata

    @tr.profile("get")
    def get(self, idx):
        """
        Get data with a given index
        """
        if self.preflight:
            ## Preflight option: This is experimental.
            ## Collect only the indeces of data to load. "populate" will load the data later.
            self.preflight_list.append(idx)
            return torch_geometric.data.Data()
        if idx in self.cache:
            ## Load data from cached buffer
            data_object = self.cache[idx]
        else:
            data_object = torch_geometric.data.Data()
            for k in self.keys:
                shape = self.vars["%s/%s" % (self.label, k)]["Shape"]
                ishape = [int(x.strip(",")) for x in shape.strip().split()]
                start = [
                    0,
                ] * len(ishape)
                count = ishape
                vdim = self.variable_dim[k]
                start[vdim] = self.variable_offset[k][idx]
                count[vdim] = self.variable_count[k][idx]
                if self.preload or self.shmem:
                    ## Read from memory when preloaded or used with shmem
                    slice_list = list()
                    for n0, n1 in zip(start, count):
                        slice_list.append(slice(n0, n0 + n1))
                    val = self.data[k][tuple(slice_list)]
                elif self.ddstore:
                    vname = "%s/%s" % (self.label, k)
                    vartype = self.vars["%s/%s" % (self.label, k)]["Type"]
                    if vartype == "double":
                        dtype = np.float64
                    elif vartype == "float":
                        dtype = np.float32
                    elif vartype == "int32_t":
                        dtype = np.int32
                    elif vartype == "int64_t":
                        dtype = np.int64
                    else:
                        raise ValueError(vartype)

                    val = np.zeros(count, dtype=dtype)
                    ## vdim should be the first dim for DDStore
                    if vdim > 0:
                        val = np.moveaxis(val, vdim, 0)
                        val = np.ascontiguousarray(val)

                    offset = start[vdim]
                    self.ddstore.get(vname, val, offset)
                    if vdim > 0:
                        val = np.moveaxis(val, 0, vdim)
                        val = np.ascontiguousarray(val)
                else:
                    ## Reading data directly from disk
                    # log("getitem out-of-memory:", self.label, k, idx)
                    val = self.f.read("%s/%s" % (self.label, k), start, count)

                v = torch.tensor(val)
                exec("data_object.%s = v" % (k))
            if self.enable_cache:
                self.cache[idx] = data_object
        return data_object

    def unlink(self):
        """
        Unlink shmem link
        """
        if self.shmem:
            for k in self.keys:
                self.shm[k].close()
                if self.local_rank == 0:
                    self.shm[k].unlink()

    def __del__(self):
        if self.ddstore:
            self.ddstore.free()
        if not self.preload and not self.shmem:
            self.f.close()
        try:
            self.unlink(self)
        except:
            pass

    def populate(self):
        """
        Populate data when preflight is on
        """
        dn = 2_000_000
        self._data = dict()
        for i in range(0, self.ndata, dn):
            for k in self.keys:
                shape = self.vars["%s/%s" % (self.label, k)]["Shape"]
                ishape = [int(x.strip(",")) for x in shape.strip().split()]
                start = [
                    0,
                ] * len(ishape)
                count = list(ishape)
                vdim = self.variable_dim[k]
                start[vdim] = self.variable_offset[k][i]
                count[vdim] = self.variable_count[k][i : i + dn].sum()

                with ad2.open(self.filename, "r", MPI.COMM_SELF) as f:
                    f.__next__()
                    self._data[k] = f.read("%s/%s" % (self.label, k), start, count)

            filtered = filter(lambda x: x >= i and x < i + dn, self.preflight_list)
            for idx in iterate_tqdm(sorted(filtered), 2, desc="AdiosWriter populate"):
                data_object = torch_geometric.data.Data()
                for k in self.keys:
                    shape = self.vars["%s/%s" % (self.label, k)]["Shape"]
                    ishape = [int(x.strip(",")) for x in shape.strip().split()]
                    start = [
                        0,
                    ] * len(ishape)
                    count = list(ishape)
                    vdim = self.variable_dim[k]
                    start[vdim] = (
                        self.variable_offset[k][idx] - self.variable_offset[k][i]
                    )
                    count[vdim] = self.variable_count[k][idx]
                    slice_list = list()
                    for n0, n1 in zip(start, count):
                        slice_list.append(slice(n0, n0 + n1))
                    val = self._data[k][tuple(slice_list)]

                    v = torch.tensor(val)
                    exec("data_object.%s = v" % (k))
                self.cache[idx] = data_object

        for k in self.keys:
            del self._data[k]

        self.preflight = False



##############################################################################
# Copyright (c) 2021, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of HydraGNN and is distributed under a BSD 3-clause      #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# SPDX-License-Identifier: BSD-3-Clause                                      #
##############################################################################

import torch
from torch.nn import ModuleList, Sequential, ReLU, Linear, Module
import torch.nn.functional as F
from torch_geometric.nn import global_mean_pool, BatchNorm
from torch.nn import GaussianNLLLoss
from hydragnn.utils.model import activation_function_selection, loss_function_selection
import sys
from hydragnn.utils.distributed import get_device


class Base(Module):
    def __init__(
        self,
        input_dim: int,
        hidden_dim: int,
        output_dim: list,
        output_type: list,
        config_heads: dict,
        activation_function_type: str,
        loss_function_type: str,
        equivariance: bool,
        ilossweights_hyperp: int = 1,  # if =1, considering weighted losses for different tasks and treat the weights as hyper parameters
        loss_weights: list = [1.0, 1.0, 1.0],  # weights for losses of different tasks
        ilossweights_nll: int = 0,  # if =1, using the scalar uncertainty as weights, as in paper# https://openaccess.thecvf.com/content_cvpr_2018/papers/Kendall_Multi-Task_Learning_Using_CVPR_2018_paper.pdf
        freeze_conv=False,
        initial_bias=None,
        dropout: float = 0.25,
        num_conv_layers: int = 16,
        num_nodes: int = None,
    ):
        super().__init__()
        self.device = get_device()
        self.input_dim = input_dim
        self.hidden_dim = hidden_dim
        self.dropout = dropout
        self.num_conv_layers = num_conv_layers
        self.graph_convs = ModuleList()
        self.feature_layers = ModuleList()
        self.num_nodes = num_nodes
        ##One head represent one variable
        ##Head can have different sizes, head_dims
        self.heads_NN = ModuleList()
        self.config_heads = config_heads
        self.head_type = output_type
        self.head_dims = output_dim
        self.num_heads = len(self.head_dims)
        ##convolutional layers for node level predictions
        self.convs_node_hidden = ModuleList()
        self.batch_norms_node_hidden = ModuleList()
        self.convs_node_output = ModuleList()
        self.batch_norms_node_output = ModuleList()
        self.equivariance = equivariance
        self.activation_function = activation_function_selection(
            activation_function_type
        )

        self.loss_function = loss_function_selection(loss_function_type)
        self.ilossweights_nll = ilossweights_nll
        self.ilossweights_hyperp = ilossweights_hyperp
        if self.ilossweights_hyperp * self.ilossweights_nll == 1:
            raise ValueError(
                "ilossweights_hyperp and ilossweights_nll cannot be both set to 1."
            )
        if self.ilossweights_hyperp == 1:
            if len(loss_weights) != self.num_heads:
                raise ValueError(
                    "Inconsistent number of loss weights and tasks: "
                    + str(len(loss_weights))
                    + " VS "
                    + str(self.num_heads)
                )
            else:
                self.loss_weights = loss_weights
            weightabssum = sum(abs(number) for number in self.loss_weights)
            self.loss_weights = [iw / weightabssum for iw in self.loss_weights]

        # Condition to pass edge_attr through forward propagation.
        self.use_edge_attr = False
        if (
            hasattr(self, "edge_dim")
            and self.edge_dim is not None
            and self.edge_dim > 0
        ):
            self.use_edge_attr = True

        # Option to only train final property layers.
        self.freeze_conv = freeze_conv
        # Option to set initially large output bias (UQ).
        self.initial_bias = initial_bias

        self._init_conv()
        if self.freeze_conv:
            self._freeze_conv()
        self._multihead()
        if self.initial_bias is not None:
            self._set_bias()

    def _init_conv(self):
        self.graph_convs.append(self.get_conv(self.input_dim, self.hidden_dim))
        self.feature_layers.append(BatchNorm(self.hidden_dim))
        for _ in range(self.num_conv_layers - 1):
            conv = self.get_conv(self.hidden_dim, self.hidden_dim)
            self.graph_convs.append(conv)
            self.feature_layers.append(BatchNorm(self.hidden_dim))

    def _conv_args(self, data):
        conv_args = {"edge_index": data.edge_index.to(torch.long)}
        if self.use_edge_attr:
            assert (
                data.edge_attr is not None
            ), "Data must have edge attributes if use_edge_attributes is set."
            conv_args.update({"edge_attr": data.edge_attr})
        return conv_args

    def _freeze_conv(self):
        for module in [self.graph_convs, self.feature_layers]:
            for layer in module:
                for param in layer.parameters():
                    param.requires_grad = False

    def _set_bias(self):
        for head, type in zip(self.heads_NN, self.head_type):
            # FIXME: we only currently enable this for graph outputs.
            if type == "graph":
                # Set the bias of the last linear layer to a large value (UQ)
                head[-1].bias.data.fill_(self.initial_bias)

    def _init_node_conv(self):
        # *******convolutional layers for node level predictions*******#
        # two ways to implement node features from here:
        # 1. one graph for all node features
        # 2. one graph for one node features (currently implemented)
        if (
            "node" not in self.config_heads
            or self.config_heads["node"]["type"] != "conv"
        ):
            return
        node_feature_ind = [
            i for i, head_type in enumerate(self.head_type) if head_type == "node"
        ]
        if len(node_feature_ind) == 0:
            return
        # In this part, each head has same number of convolutional layers, but can have different output dimension
        self.convs_node_hidden.append(
            self.get_conv(self.hidden_dim, self.hidden_dim_node[0])
        )
        self.batch_norms_node_hidden.append(BatchNorm(self.hidden_dim_node[0]))
        for ilayer in range(self.num_conv_layers_node - 1):
            self.convs_node_hidden.append(
                self.get_conv(
                    self.hidden_dim_node[ilayer], self.hidden_dim_node[ilayer + 1]
                )
            )
            self.batch_norms_node_hidden.append(
                BatchNorm(self.hidden_dim_node[ilayer + 1])
            )
        for ihead in node_feature_ind:
            self.convs_node_output.append(
                self.get_conv(self.hidden_dim_node[-1], self.head_dims[ihead])
            )
            self.batch_norms_node_output.append(BatchNorm(self.head_dims[ihead]))

    def _multihead(self):
        ############multiple heads/taks################
        # shared dense layers for heads with graph level output
        dim_sharedlayers = 0
        if "graph" in self.config_heads:
            denselayers = []
            dim_sharedlayers = self.config_heads["graph"]["dim_sharedlayers"]
            denselayers.append(Linear(self.hidden_dim, dim_sharedlayers))
            denselayers.append(self.activation_function)
            for ishare in range(self.config_heads["graph"]["num_sharedlayers"] - 1):
                denselayers.append(Linear(dim_sharedlayers, dim_sharedlayers))
                denselayers.append(self.activation_function)
            self.graph_shared = Sequential(*denselayers)

        if "node" in self.config_heads:
            self.num_conv_layers_node = self.config_heads["node"]["num_headlayers"]
            self.hidden_dim_node = self.config_heads["node"]["dim_headlayers"]
            self._init_node_conv()

        inode_feature = 0
        for ihead in range(self.num_heads):
            # mlp for each head output
            if self.head_type[ihead] == "graph":
                num_head_hidden = self.config_heads["graph"]["num_headlayers"]
                dim_head_hidden = self.config_heads["graph"]["dim_headlayers"]
                denselayers = []
                denselayers.append(Linear(dim_sharedlayers, dim_head_hidden[0]))
                denselayers.append(self.activation_function)
                for ilayer in range(num_head_hidden - 1):
                    denselayers.append(
                        Linear(dim_head_hidden[ilayer], dim_head_hidden[ilayer + 1])
                    )
                    denselayers.append(self.activation_function)
                denselayers.append(
                    Linear(
                        dim_head_hidden[-1],
                        self.head_dims[ihead] + self.ilossweights_nll * 1,
                    )
                )
                head_NN = Sequential(*denselayers)
            elif self.head_type[ihead] == "node":
                self.node_NN_type = self.config_heads["node"]["type"]
                head_NN = ModuleList()
                if self.node_NN_type == "mlp" or self.node_NN_type == "mlp_per_node":
                    self.num_mlp = 1 if self.node_NN_type == "mlp" else self.num_nodes
                    assert (
                        self.num_nodes is not None
                    ), "num_nodes must be positive integer for MLP"
                    # """if different graphs in the dataset have different size, one MLP is shared across all nodes """
                    head_NN = MLPNode(
                        self.hidden_dim,
                        self.head_dims[ihead],
                        self.num_mlp,
                        self.hidden_dim_node,
                        self.config_heads["node"]["type"],
                        self.activation_function,
                    )
                elif self.node_NN_type == "conv":
                    for conv, batch_norm in zip(
                        self.convs_node_hidden, self.batch_norms_node_hidden
                    ):
                        head_NN.append(conv)
                        head_NN.append(batch_norm)
                    head_NN.append(self.convs_node_output[inode_feature])
                    head_NN.append(self.batch_norms_node_output[inode_feature])
                    inode_feature += 1
                else:
                    raise ValueError(
                        "Unknown head NN structure for node features"
                        + self.node_NN_type
                        + "; currently only support 'mlp', 'mlp_per_node' or 'conv' (can be set with config['NeuralNetwork']['Architecture']['output_heads']['node']['type'], e.g., ./examples/ci_multihead.json)"
                    )
            else:
                raise ValueError(
                    "Unknown head type"
                    + self.head_type[ihead]
                    + "; currently only support 'graph' or 'node'"
                )
            self.heads_NN.append(head_NN)

    def forward(self, data):
        x = data.x
        pos = data.pos

        ### encoder part ####
        conv_args = self._conv_args(data)
        for conv, feat_layer in zip(self.graph_convs, self.feature_layers):
            c, pos = conv(x=x, pos=pos, **conv_args)
            x = self.activation_function(feat_layer(c))

        #### multi-head decoder part####
        # shared dense layers for graph level output
        if data.batch is None:
            x_graph = x.mean(dim=0, keepdim=True)
        else:
            x_graph = global_mean_pool(x, data.batch.to(x.device))
        outputs = []
        for head_dim, headloc, type_head in zip(
            self.head_dims, self.heads_NN, self.head_type
        ):
            if type_head == "graph":
                x_graph_head = self.graph_shared(x_graph)
                outputs.append(headloc(x_graph_head))
            else:
                if self.node_NN_type == "conv":
                    for conv, batch_norm in zip(headloc[0::2], headloc[1::2]):
                        x_node = self.activation_function(
                            batch_norm(conv(x=x, edge_index=data.edge_index))
                        )
                else:
                    x_node = headloc(x=x, batch=data.batch)
                outputs.append(x_node)
        return outputs

    def loss(self, pred, value, head_index):
        if self.ilossweights_nll == 1:
            return self.loss_nll(pred, value, head_index)
        elif self.ilossweights_hyperp == 1:
            return self.loss_hpweighted(pred, value, head_index)

    def loss_nll(self, pred, value, head_index):
        # negative log likelihood loss
        # uncertainty to weigh losses in https://openaccess.thecvf.com/content_cvpr_2018/papers/Kendall_Multi-Task_Learning_Using_CVPR_2018_paper.pdf
        # fixme: Pei said that right now this is never used
        raise ValueError("loss_nll() not ready yet")
        nll_loss = 0
        tasks_mseloss = []
        loss = GaussianNLLLoss()
        for ihead in range(self.num_heads):
            head_pre = pred[ihead][:, :-1]
            pred_shape = head_pre.shape
            head_val = value[head_index[ihead]]
            value_shape = head_val.shape
            if pred_shape != value_shape:
                head_val = torch.reshape(head_val, pred_shape)
            head_var = torch.exp(pred[ihead][:, -1])
            nll_loss += loss(head_pre, head_val, head_var)
            tasks_mseloss.append(F.mse_loss(head_pre, head_val))

        return nll_loss, tasks_mseloss, []

    def loss_hpweighted(self, pred, value, head_index):
        # weights for different tasks as hyper-parameters
        tot_loss = 0
        tasks_loss = []
        for ihead in range(self.num_heads):
            head_pre = pred[ihead]
            pred_shape = head_pre.shape
            head_val = value[head_index[ihead]]
            value_shape = head_val.shape
            if pred_shape != value_shape:
                head_val = torch.reshape(head_val, pred_shape)

            tot_loss += (
                self.loss_function(head_pre, head_val) * self.loss_weights[ihead]
            )
            tasks_loss.append(self.loss_function(head_pre, head_val))

        return tot_loss, tasks_loss

    def __str__(self):
        return "Base"


class MLPNode(Module):
    def __init__(
        self,
        input_dim,
        output_dim,
        num_mlp,
        hidden_dim_node,
        node_type,
        activation_function,
    ):
        super().__init__()
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.node_type = node_type
        self.num_mlp = num_mlp
        self.activation_function = activation_function

        self.mlp = ModuleList()
        for _ in range(self.num_mlp):
            denselayers = []
            denselayers.append(Linear(self.input_dim, hidden_dim_node[0]))
            denselayers.append(self.activation_function)
            for ilayer in range(len(hidden_dim_node) - 1):
                denselayers.append(
                    Linear(hidden_dim_node[ilayer], hidden_dim_node[ilayer + 1])
                )
                denselayers.append(self.activation_function)
            denselayers.append(Linear(hidden_dim_node[-1], output_dim))
            self.mlp.append(Sequential(*denselayers))

    def node_features_reshape(self, x, batch):
        """reshape x from [batch_size*num_nodes, num_features] to [batch_size, num_features, num_nodes]"""
        num_features = x.shape[1]
        batch_size = batch.max() + 1
        out = torch.zeros(
            (batch_size, num_features, self.num_nodes),
            dtype=x.dtype,
            device=x.device,
        )
        for inode in range(self.num_nodes):
            inode_index = [i for i in range(inode, batch.shape[0], self.num_nodes)]
            out[:, :, inode] = x[inode_index, :]
        return out

    def forward(self, x: torch.Tensor, batch: torch.Tensor):
        if self.node_type == "mlp":
            outs = self.mlp[0](x)
        else:
            outs = torch.zeros(
                (x.shape[0], self.output_dim),
                dtype=x.dtype,
                device=x.device,
            )
            x_nodes = self.node_features_reshape(x, batch)
            for inode in range(self.num_nodes):
                inode_index = [i for i in range(inode, batch.shape[0], self.num_nodes)]
                outs[inode_index, :] = self.mlp[inode](x_nodes[:, :, inode])
        return outs

    def __str__(self):
        return "MLPNode"



import os
import torch
import torch.nn.functional as F
import json

try:
    import mendeleev
except ImportError:
    pass


class atomicdescriptors:
    def __init__(
        self,
        embeddingfilename,
        overwritten=True,
        element_types=["C", "H", "O", "N", "F", "S"],
        one_hot=False,
    ):
        if os.path.exists(embeddingfilename) and not overwritten:
            print("loading from existing file: ", embeddingfilename)
            with open(embeddingfilename, "r") as f:
                self.atom_embeddings = json.load(f)
        else:
            self.atom_embeddings = {}
            if element_types is None:
                self.element_types = []
                for ele in mendeleev.get_all_elements():
                    self.element_types.append(ele.symbol)
            else:
                self.element_types = []
                for ele in mendeleev.get_all_elements():
                    if ele.symbol in element_types:
                        self.element_types.append(ele.symbol)
            self.one_hot = one_hot
            type_id = self.get_type_ids()
            group_id = self.get_group_ids()
            period = self.get_period()
            covalent_radius = self.get_covalent_radius()
            electron_affinity = self.get_electron_affinity()
            block = self.get_block()
            atomic_volume = self.get_atomic_volume()
            atomic_number = self.get_atomic_number()
            atomic_weight = self.get_atomic_weight()
            electronegativity = self.get_electronegativity()
            valenceelectrons = self.get_valence_electrons()
            ionenergies = self.get_ionenergies()
            if self.one_hot:
                # properties with integer values
                group_id = self.convert_integerproperty_onehot(group_id, num_classes=-1)
                period = self.convert_integerproperty_onehot(period, num_classes=-1)
                atomic_number = self.convert_integerproperty_onehot(
                    atomic_number, num_classes=-1
                )
                valenceelectrons = self.convert_integerproperty_onehot(
                    valenceelectrons, num_classes=-1
                )
                # properties with real values
                covalent_radius = self.convert_realproperty_onehot(
                    covalent_radius, num_classes=10
                )
                electron_affinity = self.convert_realproperty_onehot(
                    electron_affinity, num_classes=10
                )
                atomic_volume = self.convert_realproperty_onehot(
                    atomic_volume, num_classes=10
                )
                atomic_weight = self.convert_realproperty_onehot(
                    atomic_weight, num_classes=10
                )
                electronegativity = self.convert_realproperty_onehot(
                    electronegativity, num_classes=10
                )
                ionenergies = self.convert_realproperty_onehot(
                    ionenergies, num_classes=10
                )

            for iele, ele in enumerate(self.element_types):
                nfeatures = 0
                self.atom_embeddings[str(mendeleev.element(ele).atomic_number)] = []
                for var in [
                    type_id,
                    group_id,
                    period,
                    covalent_radius,
                    electron_affinity,
                    block,
                    atomic_volume,
                    atomic_number,
                    atomic_weight,
                    electronegativity,
                    valenceelectrons,
                    ionenergies,
                ]:
                    nfeatures += var.size()[1]
                    self.atom_embeddings[
                        str(mendeleev.element(ele).atomic_number)
                    ].extend(var[iele, :].tolist())
            with open(embeddingfilename, "w") as f:
                json.dump(self.atom_embeddings, f)

    def get_type_ids(self):
        type_id = F.one_hot(torch.arange(len(self.element_types)), num_classes=-1)
        return type_id

    def get_group_ids(self, num_classes=-1):
        group_id = []
        for ele in self.element_types:
            group_id.append(self.get_group_id(ele) - 1)
        return torch.Tensor(group_id).reshape(len(self.element_types), -1)

    def get_group_id(self, element):
        if mendeleev.element(element).group_id is not None:
            return mendeleev.element(element).group_id
        else:
            raise ValueError(
                f"None is returned by Mendeleev for group_id of element: {element}"
            )

    def get_period(self, num_classes=-1):
        period = []
        for ele in self.element_types:
            period.append(mendeleev.element(ele).period - 1)
        return torch.Tensor(period).reshape(len(self.element_types), -1)

    def __propertynormalize__(self, prop_list, prop_name):

        None_elements = [
            ele for ele, item in zip(self.element_types, prop_list) if item is None
        ]
        if len(None_elements) > 0:
            raise ValueError(
                f"None is identified in Mendeleev property, {prop_name}, of elements, {None_elements}"
            )

        minval = min(prop_list)
        maxval = max(prop_list)
        return [(item - minval) / (maxval - minval) for item in prop_list]

    def __realtocategorical__(self, prop_tensor, num_classes=10):

        delval = (prop_tensor.max() - prop_tensor.min()) / num_classes
        categories = torch.minimum(
            (prop_tensor - prop_tensor.min()) / delval, torch.tensor([num_classes - 1])
        )
        return categories

    def get_covalent_radius(self):
        cr = []
        for ele in self.element_types:
            cr.append(mendeleev.element(ele).covalent_radius)
        cr = self.__propertynormalize__(cr, "covalent_radius")
        return torch.Tensor(cr).reshape(len(self.element_types), -1)

    def get_atomic_number(self):
        an = []
        for ele in self.element_types:
            an.append(mendeleev.element(ele).atomic_number)
        return torch.Tensor(an).reshape(len(self.element_types), -1)

    def get_atomic_weight(self):
        aw = []
        for ele in self.element_types:
            aw.append(mendeleev.element(ele).atomic_weight)
        aw = self.__propertynormalize__(aw, "atomic_weight")
        return torch.Tensor(aw).reshape(len(self.element_types), -1)

    def get_electron_affinity(self):
        ea = []
        for ele in self.element_types:
            ea.append(mendeleev.element(ele).electron_affinity)
        ea = self.__propertynormalize__(ea, "electron_affinity")
        return torch.Tensor(ea).reshape(len(self.element_types), -1)

    def get_block(self):
        blocklist = ["s", "p", "d", "f"]
        block = []
        for ele in self.element_types:
            block.append(blocklist.index(mendeleev.element(ele).block))
        block = F.one_hot(torch.Tensor(block).long(), num_classes=-1)
        return block

    def get_atomic_volume(self):
        av = []
        for ele in self.element_types:
            av.append(mendeleev.element(ele).atomic_volume)
        av = self.__propertynormalize__(av, "atomic_volume")
        return torch.Tensor(av).reshape(len(self.element_types), -1)

    def get_electronegativity(self):
        en = []
        for ele in self.element_types:
            en.append(mendeleev.element(ele).en_pauling)
        en = self.__propertynormalize__(en, "en_pauling")
        return torch.Tensor(en).reshape(len(self.element_types), -1)

    def get_valence_electrons(self):
        ve = []
        for ele in self.element_types:
            ve.append(mendeleev.element(ele).nvalence())
        return torch.Tensor(ve).reshape(len(self.element_types), -1)

    def get_ionenergies(self):
        ie = []
        for ele in self.element_types:
            degrees = mendeleev.element(ele).ionenergies.keys()
            if len(degrees) == 0:
                ie.append(None)
            else:
                degree = min(degrees)
                ie.append(mendeleev.element(ele).ionenergies[degree])
        ie = self.__propertynormalize__(ie, "ionenergies")
        return torch.Tensor(ie).reshape(len(self.element_types), -1)

    def convert_integerproperty_onehot(self, prop, num_classes=-1):
        prop = F.one_hot(prop.to(torch.int64).squeeze(), num_classes=num_classes)
        return prop

    def convert_realproperty_onehot(self, prop, num_classes=10):
        prop = self.__realtocategorical__(prop, num_classes=num_classes)
        prop = F.one_hot(prop.to(torch.int64).squeeze(), num_classes=num_classes)
        return prop

    def get_atom_features(self, atomtype):
        if isinstance(atomtype, str):
            atomtype = mendeleev.element(atomtype).atomic_number
        return torch.tensor(self.atom_embeddings[str(atomtype)])


if __name__ == "__main__":
    atomicdescriptor = atomicdescriptors(
        "./embedding.json", overwritten=True, element_types=["C", "H", "S"]
    )
    print(atomicdescriptor.get_atom_features("C"))
    print(len(atomicdescriptor.get_atom_features("C")))
    atomicdescriptor_onehot = atomicdescriptors(
        "./embedding_onehot.json",
        overwritten=True,
        element_types=["C", "H", "S"],
        one_hot=True,
    )
    print(atomicdescriptor_onehot.get_atom_features("C"))
    print(len(atomicdescriptor_onehot.get_atom_features("C")))



##############################################################################
# Copyright (c) 2021, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of HydraGNN and is distributed under a BSD 3-clause      #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# SPDX-License-Identifier: BSD-3-Clause                                      #
##############################################################################

import os
import re

import torch
import torch.distributed as dist

from .print_utils import print_distributed

import psutil
import socket


def find_ifname(myaddr):
    """
    Find socket ifname for a given ip adress. This is for "GLOO" ddp setup.
    Usage example:
        find_ifname("127.0.0.1") will return a network interface name, such as "lo". "lo0", etc.
    """
    ipaddr = socket.gethostbyname(myaddr)
    ifname = None
    for nic, addrs in psutil.net_if_addrs().items():
        for addr in addrs:
            if addr.address == ipaddr:
                ifname = nic
                break
        if ifname is not None:
            break

    return ifname


def parse_slurm_nodelist(nodelist):
    """
    Parse SLURM_NODELIST env string to get list of nodes.
    Usage example:
        parse_slurm_nodelist(os.environ["SLURM_NODELIST"])
    Input examples:
        "or-condo-g04"
        "or-condo-g[05,07-08,13]"
        "or-condo-g[05,07-08,13],or-condo-h[01,12]"
    """
    nlist = list()
    for block, _ in re.findall(r"([\w-]+(\[[\d\-,]+\])*)", nodelist):
        m = re.match(r"^(?P<prefix>[\w\-]+)\[(?P<group>.*)\]", block)
        if m is None:
            ## single node
            nlist.append(block)
        else:
            ## multiple nodes
            g = m.groups()
            prefix = g[0]
            for sub in g[1].split(","):
                if "-" in sub:
                    start, end = re.match(r"(\d+)-(\d+)", sub).groups()
                    fmt = "%%0%dd" % (len(start))
                    for i in range(int(start), int(end) + 1):
                        node = prefix + fmt % i
                        nlist.append(node)
                else:
                    node = prefix + sub
                    nlist.append(node)

    return nlist


def init_comm_size_and_rank():
    world_size = None
    world_rank = 0

    if os.getenv("OMPI_COMM_WORLD_SIZE") and os.getenv("OMPI_COMM_WORLD_RANK"):
        ## Summit
        world_size = int(os.environ["OMPI_COMM_WORLD_SIZE"])
        world_rank = int(os.environ["OMPI_COMM_WORLD_RANK"])
    elif os.getenv("SLURM_NPROCS") and os.getenv("SLURM_PROCID"):
        ## CADES
        world_size = int(os.environ["SLURM_NPROCS"])
        world_rank = int(os.environ["SLURM_PROCID"])

    ## Fall back to default
    if world_size is None:
        world_size = 1

    return int(world_size), int(world_rank)


def get_comm_size_and_rank():
    world_size = None
    world_rank = 0

    if dist.is_initialized():
        world_size = dist.get_world_size()
        world_rank = dist.get_rank()
    else:
        world_size = 1

    return int(world_size), int(world_rank)


def setup_ddp():
    """ "Initialize DDP"""

    if os.getenv("HYDRAGNN_BACKEND") is not None:
        backend = os.environ["HYDRAGNN_BACKEND"]
    elif dist.is_nccl_available() and torch.cuda.is_available():
        backend = "nccl"
    elif torch.distributed.is_gloo_available():
        backend = "gloo"
    else:
        raise RuntimeError("No parallel backends available")

    world_size, world_rank = init_comm_size_and_rank()

    ## Default setting
    master_addr = "127.0.0.1"
    master_port = "8889"

    if os.getenv("LSB_HOSTS") is not None:
        ## source: https://www.olcf.ornl.gov/wp-content/uploads/2019/12/Scaling-DL-on-Summit.pdf
        ## The following is Summit specific
        master_addr = os.environ["LSB_HOSTS"].split()[1]
    elif os.getenv("LSB_MCPU_HOSTS") is not None:
        master_addr = os.environ["LSB_MCPU_HOSTS"].split()[2]
    elif os.getenv("SLURM_NODELIST") is not None:
        ## The following is CADES specific
        master_addr = parse_slurm_nodelist(os.environ["SLURM_NODELIST"])[0]

    try:
        if backend in ["nccl", "gloo"]:
            os.environ["MASTER_ADDR"] = master_addr
            os.environ["MASTER_PORT"] = master_port
            os.environ["WORLD_SIZE"] = str(world_size)
            os.environ["RANK"] = str(world_rank)

        if (backend == "gloo") and ("GLOO_SOCKET_IFNAME" not in os.environ):
            ifname = find_ifname(master_addr)
            if ifname is not None:
                os.environ["GLOO_SOCKET_IFNAME"] = ifname

        print_distributed(
            1,
            "Distributed data parallel: %s master at %s:%s"
            % (backend, master_addr, master_port),
        )

        if not dist.is_initialized():
            dist.init_process_group(backend=backend, init_method="env://")

    except KeyError:
        print("DDP has to be initialized within a job - Running in sequential mode")

    return world_size, world_rank


def get_device_list():

    available_gpus = [i for i in range(torch.cuda.device_count())]

    return available_gpus


def get_device_name(use_gpu=True, rank_per_model=1, verbosity_level=0):

    available_gpus = get_device_list()
    if not use_gpu or not available_gpus:
        print_distributed(verbosity_level, "Using CPU")
        return "cpu"

    world_size, world_rank = get_comm_size_and_rank()
    if rank_per_model != 1:
        raise ValueError("Exactly 1 rank per device currently supported")

    print_distributed(verbosity_level, "Using GPU")
    ## We need to ge a local rank if there are multiple GPUs available.
    localrank = 0
    if torch.cuda.device_count() > 1:
        if os.getenv("OMPI_COMM_WORLD_LOCAL_RANK"):
            ## Summit
            localrank = int(os.environ["OMPI_COMM_WORLD_LOCAL_RANK"])
        elif os.getenv("SLURM_LOCALID"):
            ## CADES
            localrank = int(os.environ["SLURM_LOCALID"])

        if localrank >= torch.cuda.device_count():
            print(
                "WARN: localrank is greater than the available device count - %d %d"
                % (localrank, torch.cuda.device_count())
            )

    device_name = "cuda:" + str(localrank)

    return device_name


def get_device_from_name(name: str):

    return torch.device(name)


def get_device(use_gpu=True, rank_per_model=1, verbosity_level=0):

    name = get_device_name(use_gpu, rank_per_model, verbosity_level)
    return get_device_from_name(name)


def is_model_distributed(model):
    return isinstance(model, torch.nn.parallel.distributed.DistributedDataParallel)


def get_distributed_model(model, verbosity=0, sync_batch_norm=False):
    device_name = get_device_name(verbosity_level=verbosity)

    if dist.is_initialized():
        if device_name == "cpu":
            model = torch.nn.parallel.DistributedDataParallel(model)
        else:
            if sync_batch_norm:
                model = torch.nn.SyncBatchNorm.convert_sync_batchnorm(model)
            device = get_device_from_name(device_name)
            model = torch.nn.parallel.DistributedDataParallel(
                model, device_ids=[device]
            )
    return model


def print_peak_memory(verbosity_level, prefix):
    # FIXME: this will have to change when the code can run on AMD gpus
    if torch.cuda.is_available():
        device = get_device()
        print_distributed(
            verbosity_level,
            f"{prefix}: {torch.cuda.max_memory_allocated(device) // 1e6}MB ",
        )


def nsplit(a, n):
    k, m = divmod(len(a), n)
    return (a[i * k + min(i, m) : (i + 1) * k + min(i + 1, m)] for i in range(n))


def comm_reduce(x, op):
    """
    All-reduce with numpy array
    """
    tx = torch.tensor(x, requires_grad=True).to(get_device())
    torch.distributed.all_reduce(tx, op=op)
    y = tx.detach().cpu().numpy()
    return y



##############################################################################
# Copyright (c) 2021, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of HydraGNN and is distributed under a BSD 3-clause      #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# SPDX-License-Identifier: BSD-3-Clause                                      #
##############################################################################

import sys, os, json, numpy as np
import pytest

import torch
from torch_geometric.data import Data
from hydragnn.preprocess.utils import (
    get_radius_graph_config,
    get_radius_graph_pbc_config,
)

from ase import build


def unittest_periodic_boundary_conditions(
    config, data, expected_neighbors, expected_neighbors_self_loops
):
    compute_edges = get_radius_graph_config(config["Architecture"], loop=False)
    compute_edges_pbc_no_self_loops = get_radius_graph_pbc_config(
        config["Architecture"], loop=False
    )
    compute_edges_pbc_with_self_loops = get_radius_graph_pbc_config(
        config["Architecture"], loop=True
    )
    num_nodes = data.pos.size(0)

    data_periodic_no_self_loops = data.clone()
    data_periodic_with_self_loops = data.clone()

    data = compute_edges(data)

    data_periodic_no_self_loops = compute_edges_pbc_no_self_loops(
        data_periodic_no_self_loops
    )
    data_periodic_with_self_loops = compute_edges_pbc_with_self_loops(
        data_periodic_with_self_loops
    )

    # Check number of nodes is unchanged.
    assert data_periodic_no_self_loops.pos.size(0) == num_nodes
    assert data_periodic_with_self_loops.pos.size(0) == num_nodes

    # Check that each node has the expected number of neighbors (with or without self interactions)
    assert (
        data_periodic_no_self_loops.edge_index.size(1) == expected_neighbors * num_nodes
    )
    assert (
        data_periodic_with_self_loops.edge_index.size(1)
        == expected_neighbors_self_loops * num_nodes
    )

    # Check the nodes were not modified.
    for i in range(num_nodes):
        for d in range(3):
            assert data_periodic_no_self_loops.pos[i][d] == data.pos[i][d]
            assert data_periodic_with_self_loops.pos[i][d] == data.pos[i][d]
        assert data_periodic_no_self_loops.x[i][0] == data.x[i][0]
        assert data_periodic_with_self_loops.x[i][0] == data.x[i][0]
    assert data_periodic_no_self_loops.y == data.y
    assert data_periodic_with_self_loops.y == data.y

    # FIXME: check lengths are at least reasonable
    for n in range(expected_neighbors * num_nodes):
        assert data_periodic_no_self_loops.edge_attr[n] < 5.0
    # Test could also check each periodic shift


def pytest_periodic_h2():
    config_file = "./tests/inputs/ci_periodic.json"
    with open(config_file, "r") as f:
        config = json.load(f)

    # Create # Hydrogen molecule (H2) with arbitrary node features
    data = Data()
    data.supercell_size = torch.tensor(
        [[3.0, 0.0, 0.0], [0.0, 3.0, 0.0], [0.0, 0.0, 3.0]]
    )
    data.atom_types = [1, 1]
    data.pos = torch.tensor([[1.0, 1.0, 1.0], [1.43, 1.43, 1.43]])
    data.x = torch.tensor([[3, 5, 7], [9, 11, 13]])
    # Create arbitrary global feature
    data.y = torch.tensor([[99]])

    # Without self loops only 1 bond per atom; with self loops each interacts with itself and its neighbor.
    unittest_periodic_boundary_conditions(config, data, 1, 2)


def pytest_periodic_bcc_large():
    config_file = "./tests/inputs/ci_periodic.json"
    with open(config_file, "r") as f:
        config = json.load(f)

    # Create BCC Chromium lattice
    config["Architecture"]["radius"] = 5.0
    unitcell = build.bulk("Cr", crystalstructure="bcc", a=3.6, orthorhombic=True)
    supercell = build.make_supercell(unitcell, np.identity(3) * 5)
    # io.write("tmp.xyz", supercell)

    # Convert to PyG
    data2 = Data()
    data2.supercell_size = torch.tensor(supercell.cell[:])
    data2.atom_types = np.ones(len(supercell)) * 27
    data2.pos = torch.tensor(supercell.positions)
    data2.x = torch.randn(data2.pos.size(0), 1)

    # Create arbitrary global feature
    data2.y = torch.tensor([[99]])

    # With this radius and BCC structure, we should get first and second shell neighbors.
    neigh_per_atom = 14
    unittest_periodic_boundary_conditions(
        config, data2, neigh_per_atom, neigh_per_atom + 1
    )



import sys
import torch
import pickle, csv

#########################################################
from rdkit import Chem
from rdkit.Chem.rdchem import BondType as BT
from rdkit.Chem.rdchem import HybridizationType
import torch.nn.functional as F
from torch_scatter import scatter
from torch_geometric.data import Data
import hydragnn

##################################################################################################################
##################################################################################################################


def get_node_attribute_name(types):
    atom_attr_name = ["atom" + k for k in types]
    extra_attr_name = [
        "atomicnumber",
        "IsAromatic",
        "HSP",
        "HSP2",
        "HSP3",
        "Hprop",
    ]
    name_list = atom_attr_name + extra_attr_name
    dims_list = [
        1,
    ] * len(name_list)
    return name_list, dims_list


def generate_graphdata_from_smilestr(simlestr, ytarget, types, var_config=None):

    ps = Chem.SmilesParserParams()
    ps.removeHs = False

    mol = Chem.MolFromSmiles(simlestr, ps)  # , sanitize=False , removeHs=False)

    data = generate_graphdata_from_rdkit_molecule(
        mol, ytarget, types, var_config=var_config
    )

    return data


def generate_graphdata_from_rdkit_molecule(
    mol, ytarget, types, atomicdescriptors_torch_tensor=None, var_config=None
):
    bonds = {BT.SINGLE: 0, BT.DOUBLE: 1, BT.TRIPLE: 2, BT.AROMATIC: 3}

    mol = Chem.AddHs(mol)
    N = mol.GetNumAtoms()

    type_idx = []
    atomic_number = []
    aromatic = []
    sp = []
    sp2 = []
    sp3 = []
    for atom in mol.GetAtoms():
        type_idx.append(types[atom.GetSymbol()])
        atomic_number.append(atom.GetAtomicNum())
        aromatic.append(1 if atom.GetIsAromatic() else 0)
        hybridization = atom.GetHybridization()
        sp.append(1 if hybridization == HybridizationType.SP else 0)
        sp2.append(1 if hybridization == HybridizationType.SP2 else 0)
        sp3.append(1 if hybridization == HybridizationType.SP3 else 0)

    z = torch.tensor(atomic_number, dtype=torch.long)

    row, col, edge_type = [], [], []
    for bond in mol.GetBonds():
        start, end = bond.GetBeginAtomIdx(), bond.GetEndAtomIdx()
        row += [start, end]
        col += [end, start]
        edge_type += 2 * [bonds[bond.GetBondType()]]

    edge_index = torch.tensor([row, col], dtype=torch.long)
    edge_type = torch.tensor(edge_type, dtype=torch.long)
    edge_attr = F.one_hot(edge_type, num_classes=len(bonds)).to(torch.float)

    perm = (edge_index[0] * N + edge_index[1]).argsort()
    edge_index = edge_index[:, perm]
    edge_attr = edge_attr[perm]

    row, col = edge_index
    hs = (z == 1).to(torch.float)
    num_hs = scatter(hs[row], col, dim_size=N).tolist()

    x1 = F.one_hot(torch.tensor(type_idx), num_classes=len(types))

    x2 = (
        torch.tensor([atomic_number, aromatic, sp, sp2, sp3, num_hs], dtype=torch.float)
        .t()
        .contiguous()
    )

    x = torch.cat([x1.to(torch.float), x2], dim=-1)

    if atomicdescriptors_torch_tensor is not None:
        assert (
            atomicdescriptors_torch_tensor.shape[0] == x.shape[0]
        ), "tensor of atomic descriptors MUST have the number of rows equal to the number of atoms in the molecule"
        x = torch.cat([x, atomicdescriptors_torch_tensor], dim=-1).to(torch.float)

    y = ytarget  # .squeeze()

    data = Data(x=x, edge_index=edge_index, edge_attr=edge_attr, y=y)
    if var_config is not None:
        hydragnn.preprocess.update_predicted_values(
            var_config["type"],
            var_config["output_index"],
            var_config["graph_feature_dims"],
            var_config["input_node_feature_dims"],
            data,
        )

    return data



import os
import numpy as np
import random

import torch
from torch import tensor
from torch_geometric.data import Data
from torch_geometric.transforms import (
    Distance,
    NormalizeRotation,
    Spherical,
    PointPairFeatures,
)

from hydragnn.utils import nsplit, tensor_divide, comm_reduce
from hydragnn.utils.print_utils import print_distributed, iterate_tqdm, log
from hydragnn.utils.distributed import get_device
from hydragnn.utils.abstractbasedataset import AbstractBaseDataset
from hydragnn.preprocess.utils import (
    get_radius_graph,
    get_radius_graph_pbc,
    get_radius_graph_config,
    get_radius_graph_pbc_config,
)
from hydragnn.preprocess import (
    update_predicted_values,
    update_atom_features,
    stratified_sampling,
)

from sklearn.model_selection import StratifiedShuffleSplit

from hydragnn.preprocess.dataset_descriptors import AtomFeatures

from abc import ABC, abstractmethod


class AbstractRawDataset(AbstractBaseDataset, ABC):
    """Raw dataset class"""

    def __init__(self, config, dist=False, sampling=None):
        super().__init__()

        """
        config:
          shows the dataset path the target variables information, e.g, location and dimension, in data file
        ###########
        dataset_list:
          list of datasets read from self.path_dictionary
        serial_data_name_list:
          list of pkl file names
        node_feature_dim:
          list of dimensions of node features
        node_feature_col:
          list of column location/index (start location if dim>1) of node features
        graph_feature_dim:
          list of dimensions of graph features
        graph_feature_col: list,
          list of column location/index (start location if dim>1) of graph features

        dist: True if RawDataLoder is distributed (i.e., each RawDataLoader will read different subset of data)
        """

        # self.serial_data_name_list = []
        self.normalize_features = (
            config["Dataset"]["normalize_features"]
            if config["Dataset"]["normalize_features"] is not None
            else False
        )
        self.node_feature_name = (
            config["Dataset"]["node_features"]["name"]
            if config["Dataset"]["node_features"]["name"] is not None
            else None
        )
        self.node_feature_dim = config["Dataset"]["node_features"]["dim"]
        self.node_feature_col = config["Dataset"]["node_features"]["column_index"]
        self.graph_feature_name = (
            config["Dataset"]["graph_features"]["name"]
            if config["Dataset"]["graph_features"]["name"] is not None
            else None
        )
        self.graph_feature_dim = config["Dataset"]["graph_features"]["dim"]
        self.graph_feature_col = config["Dataset"]["graph_features"]["column_index"]
        self.raw_dataset_name = config["Dataset"]["name"]
        self.data_format = config["Dataset"]["format"]
        self.path_dictionary = config["Dataset"]["path"]

        assert len(self.node_feature_name) == len(self.node_feature_dim)
        assert len(self.node_feature_name) == len(self.node_feature_col)
        assert len(self.graph_feature_name) == len(self.graph_feature_dim)
        assert len(self.graph_feature_name) == len(self.graph_feature_col)

        self.sampling = sampling
        self.dist = dist
        if self.dist:
            assert torch.distributed.is_initialized()
            self.world_size = torch.distributed.get_world_size()
            self.rank = torch.distributed.get_rank()

        self.__load_raw_data()

        ## SerializedDataLoader
        self.verbosity = config["Verbosity"]["level"]
        self.rotational_invariance = config["Dataset"]["rotational_invariance"]
        self.periodic_boundary_conditions = config["NeuralNetwork"]["Architecture"][
            "periodic_boundary_conditions"
        ]
        self.radius = config["NeuralNetwork"]["Architecture"]["radius"]
        self.max_neighbours = config["NeuralNetwork"]["Architecture"]["max_neighbours"]
        self.variables = config["NeuralNetwork"]["Variables_of_interest"]
        self.variables_type = config["NeuralNetwork"]["Variables_of_interest"]["type"]
        self.output_index = config["NeuralNetwork"]["Variables_of_interest"][
            "output_index"
        ]
        self.input_node_features = config["NeuralNetwork"]["Variables_of_interest"][
            "input_node_features"
        ]

        self.spherical_coordinates = False
        self.point_pair_features = False

        if "Descriptors" in config["Dataset"]:
            if "SphericalCoordinates" in config["Dataset"]["Descriptors"]:
                self.spherical_coordinates = config["Dataset"]["Descriptors"][
                    "SphericalCoordinates"
                ]
            if "PointPairFeatures" in config["Dataset"]["Descriptors"]:
                self.point_pair_features = config["Dataset"]["Descriptors"][
                    "PointPairFeatures"
                ]

        # Descriptors about topology of the local environment
        if self.spherical_coordinates and self.point_pair_features:
            # FIXME We need to a new function to include both spherical coordinates and point point features together as edge features
            # Each of the two transformation computes the distance between nodes, and adds it to the set of edge features
            # A naive simuktaneous utilization of both spherical coordinates and point point features together includes the distance multiple times in the edge-feature vector
            raise ValueError(
                "Spherical Coorindates and Point Pair Features cannot be used together in the current version of HydraGNN"
            )

        if self.spherical_coordinates:
            self.edge_feature_transform = Spherical(norm=False, cat=False)

        if self.point_pair_features:
            self.edge_feature_transform = PointPairFeatures(cat=False)

        self.subsample_percentage = None

        self.__build_edge()

    def __load_raw_data(self):
        """Loads the raw files from specified path, performs the transformation to Data objects and normalization of values.
        After that the serialized data is stored to the serialized_dataset directory.
        """

        # serialized_dir = os.environ["SERIALIZED_DATA_PATH"] + "/serialized_dataset"
        # if not os.path.exists(serialized_dir):
        #     os.makedirs(serialized_dir, exist_ok=True)

        for dataset_type, raw_data_path in self.path_dictionary.items():
            if not os.path.isabs(raw_data_path):
                raw_data_path = os.path.join(os.getcwd(), raw_data_path)
            if not os.path.exists(raw_data_path):
                raise ValueError("Folder not found: ", raw_data_path)

            assert (
                len(os.listdir(raw_data_path)) > 0
            ), "No data files provided in {}!".format(raw_data_path)

            filelist = sorted(os.listdir(raw_data_path))
            if self.dist:
                ## Random shuffle filelist to avoid the same test/validation set
                random.seed(43)
                random.shuffle(filelist)
                if self.sampling is not None:
                    filelist = np.random.choice(
                        filelist, int(len(filelist) * self.sampling)
                    )

                x = torch.tensor(len(filelist), requires_grad=False).to(get_device())
                y = x.clone().detach().requires_grad_(False)
                torch.distributed.all_reduce(x, op=torch.distributed.ReduceOp.MAX)
                assert x == y
                filelist = list(nsplit(filelist, self.world_size))[self.rank]
                log("local filelist", len(filelist))

            for name in iterate_tqdm(filelist, verbosity_level=2, desc="Local files"):
                if name == ".DS_Store":
                    continue
                # if the directory contains file, iterate over them
                if os.path.isfile(os.path.join(raw_data_path, name)):
                    data_object = self.transform_input_to_data_object_base(
                        filepath=os.path.join(raw_data_path, name)
                    )
                    if not isinstance(data_object, type(None)):
                        self.dataset.append(data_object)
                # if the directory contains subdirectories, explore their content
                elif os.path.isdir(os.path.join(raw_data_path, name)):
                    dir_name = os.path.join(raw_data_path, name)
                    for subname in os.listdir(dir_name):
                        if os.path.isfile(os.path.join(dir_name, subname)):
                            data_object = self.transform_input_to_data_object_base(
                                filepath=os.path.join(dir_name, subname)
                            )
                            if not isinstance(data_object, type(None)):
                                self.dataset.append(data_object)
            if self.dist:
                torch.distributed.barrier()

        # scaled features by number of nodes
        self.__scale_features_by_num_nodes()

        if self.normalize_features:
            self.__normalize_dataset()

    def __normalize_dataset(self):

        """Performs the normalization on Data objects and returns the normalized dataset."""
        num_node_features = len(self.node_feature_dim)
        num_graph_features = len(self.graph_feature_dim)

        self.minmax_graph_feature = np.full((2, num_graph_features), np.inf)
        # [0,...]:minimum values; [1,...]: maximum values
        self.minmax_node_feature = np.full((2, num_node_features), np.inf)
        self.minmax_graph_feature[1, :] *= -1
        self.minmax_node_feature[1, :] *= -1
        for data in self.dataset:
            # find maximum and minimum values for graph level features
            g_index_start = 0
            for ifeat in range(num_graph_features):
                g_index_end = g_index_start + self.graph_feature_dim[ifeat]
                self.minmax_graph_feature[0, ifeat] = min(
                    torch.min(data.y[g_index_start:g_index_end]),
                    self.minmax_graph_feature[0, ifeat],
                )
                self.minmax_graph_feature[1, ifeat] = max(
                    torch.max(data.y[g_index_start:g_index_end]),
                    self.minmax_graph_feature[1, ifeat],
                )
                g_index_start = g_index_end

            # find maximum and minimum values for node level features
            n_index_start = 0
            for ifeat in range(num_node_features):
                n_index_end = n_index_start + self.node_feature_dim[ifeat]
                self.minmax_node_feature[0, ifeat] = min(
                    torch.min(data.x[:, n_index_start:n_index_end]),
                    self.minmax_node_feature[0, ifeat],
                )
                self.minmax_node_feature[1, ifeat] = max(
                    torch.max(data.x[:, n_index_start:n_index_end]),
                    self.minmax_node_feature[1, ifeat],
                )
                n_index_start = n_index_end

        ## Gather minmax in parallel
        if self.dist:
            self.minmax_graph_feature[0, :] = comm_reduce(
                self.minmax_graph_feature[0, :], torch.distributed.ReduceOp.MIN
            )
            self.minmax_graph_feature[1, :] = comm_reduce(
                self.minmax_graph_feature[1, :], torch.distributed.ReduceOp.MAX
            )
            self.minmax_node_feature[0, :] = comm_reduce(
                self.minmax_node_feature[0, :], torch.distributed.ReduceOp.MIN
            )
            self.minmax_node_feature[1, :] = comm_reduce(
                self.minmax_node_feature[1, :], torch.distributed.ReduceOp.MAX
            )

        for data in self.dataset:
            g_index_start = 0
            for ifeat in range(num_graph_features):
                g_index_end = g_index_start + self.graph_feature_dim[ifeat]
                data.y[g_index_start:g_index_end] = tensor_divide(
                    (
                        data.y[g_index_start:g_index_end]
                        - self.minmax_graph_feature[0, ifeat]
                    ),
                    (
                        self.minmax_graph_feature[1, ifeat]
                        - self.minmax_graph_feature[0, ifeat]
                    ),
                )
                g_index_start = g_index_end
            n_index_start = 0
            for ifeat in range(num_node_features):
                n_index_end = n_index_start + self.node_feature_dim[ifeat]
                data.x[:, n_index_start:n_index_end] = tensor_divide(
                    (
                        data.x[:, n_index_start:n_index_end]
                        - self.minmax_node_feature[0, ifeat]
                    ),
                    (
                        self.minmax_node_feature[1, ifeat]
                        - self.minmax_node_feature[0, ifeat]
                    ),
                )
                n_index_start = n_index_end

    @abstractmethod
    def transform_input_to_data_object_base(self, filepath):
        pass

    def __scale_features_by_num_nodes(self):
        """Calculate [**]_scaled_num_nodes"""
        scaled_graph_feature_index = [
            i
            for i in range(len(self.graph_feature_name))
            if "_scaled_num_nodes" in self.graph_feature_name[i]
        ]
        scaled_node_feature_index = [
            i
            for i in range(len(self.node_feature_name))
            if "_scaled_num_nodes" in self.node_feature_name[i]
        ]

        for idx, data_object in enumerate(self.dataset):
            if self.dataset[idx].y is not None:
                self.dataset[idx].y[scaled_graph_feature_index] = (
                    self.dataset[idx].y[scaled_graph_feature_index]
                    / data_object.num_nodes
                )
            if self.dataset[idx].x is not None:
                self.dataset[idx].x[:, scaled_node_feature_index] = (
                    self.dataset[idx].x[:, scaled_node_feature_index]
                    / data_object.num_nodes
                )

    def __build_edge(self):
        """Loads the serialized structures data from specified path, computes new edges for the structures based on the maximum number of neighbours and radius. Additionally,
        atom and structure features are updated.

        Parameters
        ----------
        dataset_path: str
            Directory path where files containing serialized structures are stored.
        config: dict
            Dictionary containing information needed to load the data and transform it, respectively: atom_features, radius, max_num_node_neighbours and predicted_value_option.
        Returns
        ----------
        [Data]
            List of Data objects representing atom structures.
        """

        rotational_invariance = NormalizeRotation(max_points=-1, sort=False)
        if self.rotational_invariance:
            self.dataset[:] = [rotational_invariance(data) for data in self.dataset]

        if self.periodic_boundary_conditions:
            # edge lengths already added manually if using PBC, so no need to call Distance.
            compute_edges = get_radius_graph_pbc(
                radius=self.radius,
                loop=False,
                max_neighbours=self.max_neighbours,
            )
        else:
            compute_edges = get_radius_graph(
                radius=self.radius,
                loop=False,
                max_neighbours=self.max_neighbours,
            )
            compute_edge_lengths = Distance(norm=False, cat=True)

        self.dataset[:] = [compute_edges(data) for data in self.dataset]

        #################################
        #### COMPUTE EDGE ATTRIBUTES ####
        #################################

        # edge lengths already added manually if using PBC.
        # if spherical coordinates or pair point is set up, then skip directly to edge_transformation
        if (not self.periodic_boundary_conditions) and (
            not hasattr(self, "edge_feature_transform")
        ):
            self.dataset[:] = [compute_edge_lengths(data) for data in self.dataset]

            max_edge_length = torch.Tensor([float("-inf")])

            for data in self.dataset:
                max_edge_length = torch.max(max_edge_length, torch.max(data.edge_attr))

            if self.dist:
                ## Gather max in parallel
                device = max_edge_length.device
                max_edge_length = max_edge_length.to(get_device())
                torch.distributed.all_reduce(
                    max_edge_length, op=torch.distributed.ReduceOp.MAX
                )
                max_edge_length = max_edge_length.to(device)

            # Normalization of the edges
            for data in self.dataset:
                data.edge_attr = data.edge_attr / max_edge_length

        # Descriptors about topology of the local environment
        elif hasattr(self, "edge_feature_transform"):
            self.dataset[:] = [
                self.edge_feature_transform(data) for data in self.dataset
            ]

        if "subsample_percentage" in self.variables.keys():
            self.subsample_percentage = self.variables["subsample_percentage"]
            sampled = stratified_sampling(
                dataset=self.dataset, subsample_percentage=self.subsample_percentage
            )
            self.dataset = sampled

    def len(self):
        return len(self.dataset)

    def get(self, idx):
        return self.dataset[idx]



from .convert_total_energy_to_formation_gibbs import (
    convert_raw_data_energy_to_gibbs,
    compute_formation_enthalpy,
)
from .compositional_histogram_cutoff import compositional_histogram_cutoff



##############################################################################
# Copyright (c) 2021, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of HydraGNN and is distributed under a BSD 3-clause      #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# SPDX-License-Identifier: BSD-3-Clause                                      #
##############################################################################

import os
import numpy as np
import hydragnn, tests
import pytest
from utils.lsms import (
    convert_raw_data_energy_to_gibbs,
)


def unittest_formation_enthalpy():

    dir = "dataset/unit_test_enthalpy"
    if not os.path.exists(dir):
        os.makedirs(dir)

    # Create random samples.
    num_config = 10
    tests.deterministic_graph_data(
        dir,
        num_config,
        number_types=2,
        linear_only=True,
    )

    # Create pure components.
    tests.deterministic_graph_data(
        dir,
        number_configurations=1,
        configuration_start=num_config,
        number_types=1,
        types=[0],
        linear_only=True,
    )
    tests.deterministic_graph_data(
        dir,
        number_configurations=1,
        configuration_start=num_config + 1,
        number_types=1,
        types=[1],
        linear_only=True,
    )

    convert_raw_data_energy_to_gibbs(dir, [0, 1], create_plots=False)

    new_dir = dir + "_gibbs_energy"
    for filename in os.listdir(new_dir):
        path = os.path.join(new_dir, filename)
        enthalpy = np.loadtxt(path, max_rows=1)
        assert enthalpy == 0


@pytest.mark.mpi_skip()
def pytest_formation_enthalpy():
    unittest_formation_enthalpy()



import torch
from .distributed import get_device_name
from torch.distributed.optim import ZeroRedundancyOptimizer

deepspeed_available = True
try:
    import deepspeed
except ImportError:
    deepspeed_available = False


def select_standard_optimizer(model, config):
    optimizer = None

    if config["type"] == "SGD":
        optimizer = torch.optim.SGD(model.parameters(), lr=config["learning_rate"])
    elif config["type"] == "Adam":
        optimizer = torch.optim.Adam(model.parameters(), lr=config["learning_rate"])
    elif config["type"] == "Adadelta":
        optimizer = torch.optim.Adadelta(model.parameters(), lr=config["learning_rate"])
    elif config["type"] == "Adagrad":
        optimizer = torch.optim.Adagrad(model.parameters(), lr=config["learning_rate"])
    elif config["type"] == "Adamax":
        optimizer = torch.optim.Adamax(model.parameters(), lr=config["learning_rate"])
    elif config["type"] == "AdamW":
        optimizer = torch.optim.AdamW(model.parameters(), lr=config["learning_rate"])
    elif config["type"] == "RMSprop":
        optimizer = torch.optim.RMSprop(model.parameters(), lr=config["learning_rate"])
    elif config["type"] == "FusedLAMB":
        assert deepspeed_available, "deepspeed package not installed"
        assert (
            "cpu" != get_device_name()
        ), "GPUs not available to use FusedLAMB optimizer from deepspeed package"
        optimizer = deepspeed.ops.lamb.FusedLamb(
            model.parameters(), lr=config["learning_rate"]
        )
    else:
        raise NameError("The string used to identify the optimizer is NOT recognized")

    return optimizer


def select_zero_redundancy_optimizer(model, config):
    optimizer = None

    if config["type"] == "SGD":
        optimizer = ZeroRedundancyOptimizer(
            model.parameters(),
            optimizer_class=torch.optim.SGD,
            lr=config["learning_rate"],
        )
    elif config["type"] == "Adam":
        optimizer = ZeroRedundancyOptimizer(
            model.parameters(),
            optimizer_class=torch.optim.Adam,
            lr=config["learning_rate"],
        )
    elif config["type"] == "Adadelta":
        optimizer = ZeroRedundancyOptimizer(
            model.parameters(),
            optimizer_class=torch.optim.Adadelta,
            lr=config["learning_rate"],
        )
    elif config["type"] == "Adagrad":
        optimizer = ZeroRedundancyOptimizer(
            model.parameters(),
            optimizer_class=torch.optim.Adagrad,
            lr=config["learning_rate"],
        )
    elif config["type"] == "Adamax":
        optimizer = ZeroRedundancyOptimizer(
            model.parameters(),
            optimizer_class=torch.optim.Adagrad,
            lr=config["learning_rate"],
        )
    elif config["type"] == "AdamW":
        optimizer = ZeroRedundancyOptimizer(
            model.parameters(),
            optimizer_class=torch.optim.AdamW,
            lr=config["learning_rate"],
        )
    elif config["type"] == "RMSprop":
        optimizer = ZeroRedundancyOptimizer(
            model.parameters(),
            optimizer_class=torch.optim.RMSprop,
            lr=config["learning_rate"],
        )
    elif config["type"] == "FusedLAMB":
        assert deepspeed_available, "deepspeed package not installed"
        assert (
            "cpu" != get_device_name()
        ), "GPUs not available to use FusedLAMB optimizer from deepspeed package"
        optimizer = ZeroRedundancyOptimizer(
            model.parameters(),
            optimizer_class=deepspeed.ops.lamb.FusedLamb,
            lr=config["learning_rate"],
        )
    else:
        raise NameError("The string used to identify the optimizer is NOT recognized")

    return optimizer


def select_optimizer(model, config):
    use_zero = False

    if "use_zero_redundancy" in config:
        use_zero = config["use_zero_redundancy"]

    if use_zero:
        return select_zero_redundancy_optimizer(model, config)
    else:
        return select_standard_optimizer(model, config)
